"""
Description: This module provides functions to load component models and configurations.
Author: Martin Altenburger
"""

import importlib
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Type, Union, cast

from loguru import logger
from pydantic import BaseModel, Field, create_model

from encodapy.components.basic_component_config import (
    ConfigData,
    InputData,
    IOAllocationModel,
    OutputData,
)

if TYPE_CHECKING:
    from encodapy.components.basic_component import BasicComponent


class ModelTypes(Enum):
    """
    Enumeration of model types for components.
    """

    COMPONENT = "component"
    COMPONENT_CONFIG = "component_config"


def check_component_type(component_type: str):
    """
    Check the component type and return the component type and path.
    If the component type is not fully qualified, return None for the path.
    """
    if "." not in component_type:
        return component_type, None

    component_type_name = component_type.rsplit(".", 1)[-1]
    component_type_path = ".".join(
        [p.strip(".") for p in (component_type, component_type_name)]
    )
    return component_type_name, component_type_path


def get_component_model(
    component_type: str,
    model_type: Optional[ModelTypes],
    model_subname: Optional[str] = None,
    none_allowed: bool = False,
) -> Union[Type["BaseModel"], Type["BasicComponent"], None]:
    """
    Function to get the model information for the component.

    Args:
        component_type (str): Type of the component
        module_path (Optional[str], optional): Path to the module, \
            if not part of EnCoDaPy. Defaults to None.

    Returns:
        Union[Type["BaseModel"], Type["BasicComponent"], None]: \
            The model if found
    """

    component_type, module_base_path = check_component_type(component_type)

    if module_base_path is None:
        module_base_path = f"encodapy.components.{component_type}.{component_type}"

    if model_type is ModelTypes.COMPONENT:
        module_path = module_base_path
    elif model_type is ModelTypes.COMPONENT_CONFIG:
        module_path = f"{module_base_path}_config"
    else:
        raise ValueError(f"Unknown model type: {model_type}")

    config_module = importlib.import_module(module_path)
    model_name = "".join(part.capitalize() for part in component_type.split("_"))

    model_name = model_name if model_subname is None else f"{model_name}{model_subname}"

    try:
        class_model = getattr(config_module, model_name)
        return class_model
    except AttributeError:
        error_msg = f"Class model '{model_name}' not found in {config_module.__name__}"
        if none_allowed:
            logger.debug(error_msg)
            return None
        logger.error(error_msg)
    raise AttributeError(error_msg)


def get_component_class_model(component_type: str) -> "Type[BasicComponent]":
    """Get the component class model for a specific component type.

    Args:
        component_type (str): Type of the component
        module_path (Optional[str], optional): Path to the module. \
            Required if not part of EnCoDaPy. Defaults to None.

    Returns:
        Type[BasicComponent]: The component class model
    """
    module = importlib.import_module("encodapy.components.basic_component")
    basic_component_class = getattr(module, "BasicComponent")

    component_class = get_component_model(
        component_type=component_type,
        model_type=ModelTypes.COMPONENT,
    )

    if component_class is None:
        raise KeyError(f"Component class not found for {component_type}")
    if issubclass(component_class, basic_component_class):
        return cast("Type[BasicComponent]", component_class)

    error_msg = f"Component class {component_class.__name__} is not a subclass of BasicComponent"
    logger.error(error_msg)
    raise TypeError(error_msg)


def get_component_io_model(
    component_type: str, model_subname: str
) -> Union[Type[InputData], Type[OutputData]]:
    """Get the component io (input or output) model for a specific component type.

    Args:
        component_type (str): Type of the component
        model_name (Optional[str], optional): Name of the Type of the model.
        module_path (Optional[str], optional): Path to the module. \
            Required if not part of EnCoDaPy. Defaults to None.

    Returns:
        Union[Type[InputData], Type[OutputData]]: The component io model with the model subname
    """
    component_data_model = get_component_model(
        component_type=component_type,
        model_subname=model_subname,
        model_type=ModelTypes.COMPONENT_CONFIG,
    )

    if component_data_model is None:
        raise KeyError(f"Component Config Model not found for {component_type}")

    if not issubclass(component_data_model, (InputData, OutputData)):
        error_msg = (
            f"Component class {component_data_model.__name__} "
            "is not a subclass of BaseModel"
        )
        logger.error(error_msg)
        raise TypeError(error_msg)

    # 1. Create the fields for the new model
    fields: Dict[str, Tuple[Any, Any]] = {}
    for datapoint_name, datapoint_info in component_data_model.model_fields.items():
        fields[datapoint_name] = (
            (
                IOAllocationModel
                if datapoint_info.is_required()
                else Optional[IOAllocationModel]
            ),
            Field(
                default=None if not datapoint_info.is_required() else ...,
                description=datapoint_info.description,
                json_schema_extra=datapoint_info.json_schema_extra,
            ),
        )

    # 2. Create the new model dynamically
    component_config_model = create_model(
        f"IO_{component_data_model.__name__}",
        **cast(Dict[str, Any], fields),
        __base__=component_data_model,
    )
    # 3. copy methods (if any)
    for attr_name, attr_value in component_data_model.__dict__.items():
        if callable(attr_value) and not attr_name.startswith("__"):
            setattr(component_config_model, attr_name, attr_value)

    # 4. copy validators (if any)
    for name, validator in component_data_model.__dict__.items():
        if hasattr(validator, "__validator_function__"):
            setattr(component_config_model, name, validator)

    return cast(Union[Type[InputData], Type[OutputData]], component_config_model)


def get_component_data_model(
    component_type: str,
    model_subname: str,
    data_model_type: Union[Type[InputData], Type[OutputData], Type[ConfigData]],
    none_allowed=True,
) -> Union[Type[InputData], Type[OutputData], Type[ConfigData], None]:
    """
    Get the component data model for a specific component type.

    Args:
        component_type (str): Type of the component
        model_subname (str): Subname of the model
        data_model_type (Union[Type[InputData], Type[OutputData], Type[ConfigData]]): \
            Type of the data model
        none_allowed (bool, optional): Whether None is allowed as a return value. \
            Defaults to True.

    Raises:
        KeyError: If the component data model is not found.
        TypeError: If the component data model is not of the expected type.

    Returns:
        Union[Type[InputData], Type[OutputData], Type[ConfigData], None]: \
            The component data model or None if not found.
    """
    data_model = get_component_model(
        component_type=component_type,
        model_subname=model_subname,
        model_type=ModelTypes.COMPONENT_CONFIG,
        none_allowed=True,
    )

    if data_model is None:
        if none_allowed:
            logger.debug(f"No data model found for {component_type}")
            return None
        raise KeyError(f"Component Config Model not found for {component_type}")
    if not issubclass(data_model, data_model_type):
        error_message = (
            f"Component class {data_model.__name__} "
            f"is not a subclass of {data_model_type.__name__}"
        )
        logger.error(error_message)
        raise TypeError(error_message)
    return cast(Union[Type[InputData], Type[OutputData], Type[ConfigData]], data_model)


def get_component_config_data_model(
    component_type: str, model_subname: str
) -> Union[Type[ConfigData], None]:
    """Get the component config data model for a specific component type.

    Args:
        component_type (str): Type of the component
        model_subname (Optional[str], optional): Name of the Type of the model.

    Returns:
        Union[Type[BasicComponent], None]: The component static data model or None, if not found.
    """
    return cast(
        Optional[Type[ConfigData]],
        get_component_data_model(
            component_type=component_type,
            model_subname=model_subname,
            data_model_type=ConfigData,
            none_allowed=True,
        ),
    )


def get_component_input_data_model(component_type: str) -> Type[InputData]:
    """Get the component input data model for a specific component type.

    Args:
        component_type (str): Type of the component

    Returns:
        Type[InputData]: The component input data model.
    """
    return cast(
        Type[InputData],
        get_component_data_model(
            component_type=component_type,
            model_subname="InputData",
            data_model_type=InputData,
        ),
    )


def get_component_output_data_model(component_type: str) -> Type[OutputData]:
    """Get the component output data model for a specific component type.

    Args:
        component_type (str): Type of the component

    Returns:
        Type[OutputData]: The component output data model.
    """
    return cast(
        Type[OutputData],
        get_component_data_model(
            component_type=component_type,
            model_subname="OutputData",
            data_model_type=OutputData,
        ),
    )



"""
Module for the basic service class for the data processing and transfer via different interfaces.
Author: Martin Altenburger
"""

import os
import sys
from asyncio import sleep
from datetime import datetime
from typing import Union, Optional
import asyncio
from loguru import logger
from pydantic import ValidationError
from encodapy.config import (
    AttributeModel,
    CommandModel,
    ConfigModel,
    DataQueryTypes,
    DefaultEnvVariables,
    Interfaces,
    OutputModel,
)
from encodapy.service.communication import (
    FileConnection,
    FiwareConnection,
    MqttConnection,
)
from encodapy.utils.error_handling import ConfigError, InterfaceNotActive
from encodapy.utils.health import update_health_file
from encodapy.utils.logging import LoggerControl
from encodapy.utils.models import (
    DataTransferComponentModel,
    DataTransferModel,
    InputDataModel,
    OutputDataEntityModel,
    OutputDataModel,
    StaticDataEntityModel,
)
from encodapy.utils.units import get_time_unit_seconds


class ControllerBasicService(FiwareConnection, FileConnection, MqttConnection):
    """
    Class for processing the data transfer to different connections
    and start a function to do the calculations.

    """

    def __init__(self, shutdown_event: Optional[asyncio.Event] = None) -> None:
        FiwareConnection.__init__(self)
        FileConnection.__init__(self)
        MqttConnection.__init__(self)

        self.shutdown_event = shutdown_event or asyncio.Event()
        self.logger = LoggerControl()

        self.reload_staticdata = False
        self.staticdata = None

        self.timestamp_health = None

        self.prepare_basic_start()

    def _load_config(self):
        """
        Function loads the environemtal variables and the config of the service.

        """
        config_path = os.environ.get(
            "CONFIG_PATH", DefaultEnvVariables.CONFIG_PATH.value
        )

        try:
            self.config = ConfigModel.from_json(file_path=config_path)
        except (
            FileNotFoundError,
            ValidationError,
            ConfigError,
            InterfaceNotActive,
        ) as e:
            logger.error(f"Error loading configuration file: {e}")
            sys.exit(1)

        if self.config.interfaces.fiware:
            self.load_fiware_params()

        if self.config.interfaces.file:
            self.load_file_params()

        if self.config.interfaces.mqtt:
            self.load_mqtt_params()

        self.reload_staticdata = os.getenv(
            "RELOAD_STATICDATA", str(DefaultEnvVariables.RELOAD_STATICDATA.value)
        ).lower() in ("true", "1", "t")

        logger.debug("ENVs succesfully loaded.")

        return config_path

    def prepare_basic_start(self):
        """
        Function to create important objects with the configuration from the configuration
        file (.env) and prepare the start basics of the service.
        """
        logger.info("Prepare the basic start of the service")

        self._load_config()

        interfaces = getattr(self.config, "interfaces", None)
        if interfaces:
            if getattr(interfaces, "fiware", False):
                self.prepare_fiware_connection()

            if getattr(interfaces, "mqtt", False):
                self.prepare_mqtt_connection()

        # Load the static data from the configuration, \
        # maybe it is needed for the preparation of components
        try:
            self.staticdata = self.reload_static_data(
                method=DataQueryTypes.CALIBRATION, staticdata=[]
            )
        except (KeyError, ValueError, TypeError) as e:
            logger.error(f"Error reloading static data: {e}")
            self.cleanup_service()
            raise

        # Prepare the individual start of the service
        try:
            self.prepare_start()
        except (KeyError, ValueError, TypeError, ValidationError) as e:
            logger.error(f"Error preparing the start of the service: {e}")
            self.cleanup_service()
            raise

    def prepare_start(self):
        """
        Function prepare the specific aspects of the start of the service \
            Fuction is called by the function for the basic preparing

        This function can be overwritten in the specific service.
        
        The function should not be do anything time consuming, \
            because the health check is not running yet.

        """
        logger.debug("There is nothing else to prepare for the start of the service.")

    def reload_static_data(self, method: DataQueryTypes, staticdata: list) -> list:
        """
        Function to reload the static data
        Args:
            method (DataQueryTypes): method for the data query
            staticdata (list): list of static data
        Returns:
            list: list of static data
        """
        if len(self.config.staticdata) == 0:
            return []

        for static_entity in self.config.staticdata:
            if static_entity.interface == Interfaces.FIWARE:
                staticdata.append(
                    StaticDataEntityModel(
                        **self.get_data_from_fiware(
                            method=method,
                            entity=static_entity,
                            timestamp_latest_output=None,
                        ).model_dump()
                    )
                )

            if static_entity.interface == Interfaces.FILE:
                staticdata.append(
                    StaticDataEntityModel(
                        **self.get_staticdata_from_file(
                            entity=static_entity,
                        ).model_dump()
                    )
                )

            if static_entity.interface == Interfaces.MQTT:
                logger.warning("interface MQTT for staticdata not supported")

        return staticdata

    async def get_data(self, method: DataQueryTypes) -> InputDataModel:
        """
        Function to get the data of all input entities via the different interfaces
        (FIWARE, FILE, MQTT)

        Args:
            method (DataQueryTypes): Method for the data query

        Returns:
            InputDataModel: Model with the input data

        """

        input_data = []
        output_timestamps = []
        output_latest_timestamps = []

        if self.config is None:
            logger.error(
                "Configuration is not loaded. Please call prepare_start() first."
            )
            return InputDataModel(
                input_entities=[], output_entities=[], static_entities=[]
            )

        for output_entity in self.config.outputs:
            if output_entity.interface == Interfaces.FIWARE:
                entity_timestamps, output_latest_timestamp = (
                    self._get_last_timestamp_for_fiware_output(output_entity)
                )

                output_timestamps.append(entity_timestamps)
                output_latest_timestamps.append(output_latest_timestamp)

            elif output_entity.interface == Interfaces.FILE:
                entity_timestamps, output_latest_timestamp = (
                    self._get_last_timestamp_for_file_output(output_entity)
                )
                output_timestamps.append(entity_timestamps)
                output_latest_timestamps.append(output_latest_timestamp)
                logger.info("File interface, output_latest_timestamp is not defined.")

            elif output_entity.interface == Interfaces.MQTT:
                entity_timestamps, output_latest_timestamp = (
                    self._get_last_timestamp_for_mqtt_output(output_entity)
                )
                output_timestamps.append(entity_timestamps)
                output_latest_timestamps.append(output_latest_timestamp)

            await sleep(0.01)

        if None in output_latest_timestamps:
            output_latest_timestamp = None
        else:
            if len(output_latest_timestamps) > 0:
                output_latest_timestamp = min(output_latest_timestamps)
            else:
                output_latest_timestamp = None

        for input_entity in self.config.inputs:
            if input_entity.interface == Interfaces.FIWARE:
                input_data.append(
                    self.get_data_from_fiware(
                        method=method,
                        entity=input_entity,
                        timestamp_latest_output=output_latest_timestamp,
                    )
                )

            elif input_entity.interface == Interfaces.FILE:
                input_data.append(
                    self.get_data_from_file(method=method, entity=input_entity)
                )

            elif input_entity.interface == Interfaces.MQTT:
                input_data.append(
                    self.get_data_from_mqtt(
                        method=method,
                        entity=input_entity,
                    )
                )

            await sleep(0.01)

        if self.reload_staticdata or self.staticdata is None:
            self.staticdata = self.reload_static_data(method=method, staticdata=[])

        return InputDataModel(
            input_entities=input_data,
            output_entities=output_timestamps,
            static_entities=self.staticdata,
        )

    def _get_output_entity_config(
        self,
        output_entity_id: str,
    ) -> Union[OutputModel, None]:
        """
        Function to get the configuration of the output attributes

        Args:
            - output_entity: id of the output entity

        Returns:
            - Union[OutputModel, None]: configuration of the output entity
            or None if the entity is not found
        """
        for entity in self.config.outputs:
            if entity.id == output_entity_id:
                return entity

        return None

    def _get_output_attribute_config(
        self,
        output_entity_id: str,
        output_attribute_id: str,
    ) -> Union[AttributeModel, None]:
        """
        Function to get the configuration of the output attribute

        Args:
            - output_entity: id of the output entity
            - output_attribute: id of the output attribute

        Returns:
            - Union[AttributeModel, None]: configuration of the output attribute
            or None if the attribute is not found
        """
        for entity in self.config.outputs:
            if entity.id == output_entity_id:
                for attribute in entity.attributes:
                    if attribute.id == output_attribute_id:
                        return attribute

        return None

    def _get_output_command_config(
        self,
        output_entity_id: str,
        output_command_id: str,
    ) -> Union[CommandModel, None]:
        """
        Function to get the configuration of the output attribute

        Args:
            - output_entity: id of the output entity
            - output_attribute: id of the output attribute

        Returns:
            - Union[AttributeModel, None]: configuration of the output attribute
            or None if the attribute is not found
        """
        for entity in self.config.outputs:
            if entity.id == output_entity_id:
                for commmand in entity.commands:
                    if commmand.id == output_command_id:
                        return commmand

        return None

    async def send_outputs(self, data_output: Union[OutputDataModel, None]):
        """
        Send output data to the interfaces defined in the Config (FIWARE, MQTT, File)

        Args:
            - data_output: OutputDataModel with the output data

        TODO: - Implement a way to use different interfaces (MQTT)
        """

        if data_output is None:
            logger.debug("No data for sending out to  instance (FIWARE, MQTT, FILE)")
            return

        for output in data_output.entities:
            output_entity = self._get_output_entity_config(output_entity_id=output.id)
            output_attributes = []
            output_commands = []

            if output_entity is None:
                logger.debug(f"Output entity {output.id} not found in configuration.")
                continue

            for attribute in output.attributes:
                output_attribute = self._get_output_attribute_config(
                    output_entity_id=output.id, output_attribute_id=attribute.id
                )

                if output_attribute is None:
                    logger.debug(
                        f"Output attribute {attribute.id} not found in configuration."
                    )
                    continue

                output_attribute.value = attribute.value
                output_attribute.unit = attribute.unit
                output_attribute.timestamp = attribute.timestamp
                output_attributes.append(output_attribute)

            for command in output.commands:
                output_command = self._get_output_command_config(
                    output_entity_id=output.id, output_command_id=command.id
                )

                if output_command is None:
                    logger.debug(
                        f"Output attribute {command.id} not found in configuration."
                    )
                    continue

                output_command.value = command.value
                output_commands.append(output_command)

            if output_entity.interface is Interfaces.FIWARE:
                await self._send_data_to_fiware(
                    output_entity=output_entity,
                    output_attributes=output_attributes,
                    output_commands=output_commands,
                )

            elif output_entity.interface is Interfaces.FILE:
                self.send_data_to_json_file(
                    output_entity=output_entity,
                    output_attributes=output_attributes,
                    output_commands=output_commands,
                )

            elif output_entity.interface is Interfaces.MQTT:
                self.send_data_to_mqtt(
                    output_entity=output_entity,
                    output_attributes=output_attributes,
                )

            await sleep(0.01)

        logger.debug("Finished sending output data")

    async def _hold_sampling_time(
        self, start_time: datetime, hold_time: Union[int, float]
    ):
        """
        Wait in each cycle until the sampling time (or cycle time) is up. If the algorithm takes
        more time than the sampling time, a warning will be given.
        Args:
            start_time: datetime, start time of the cycle
            hold_time: int or float, sampling time in seconds
        """
        if ((datetime.now() - start_time).total_seconds()) > hold_time:
            logger.warning(
                "The processing time is longer than the sampling time."
                "The sampling time must be increased!"
            )
        while ((datetime.now() - start_time).total_seconds()) < hold_time:
            if self.shutdown_event.is_set():
                break
            await sleep(0.01)

    async def calculation(
        self,
        data: InputDataModel,
    ) -> Union[DataTransferModel, None]:
        """
        Function to start the calculation, do something with data - used in the services
        Only a dummy function, has to be implemented in the services
        Args:
            - data: InputDataModel with the input data
        Returns:
            - Union[DataTransferModel, None]: Output data from the calculation
        """
        logger.debug(
            "No calculation function implemented, "
            f"get the data only: {data.model_dump_json()}"
        )
        return None

    async def calibration(self, data: InputDataModel):
        """
        Function to start the calibration, do something with data - used in the services
        Only a dummy function, has to be implemented in the services
        """

        logger.debug(
            "No calibration function implemented, "
            f"get the data only: {data.model_dump_json()}"
        )
        return None

    def prepare_output(self, data_output: DataTransferModel) -> OutputDataModel:
        """
        Function to prepare the output data for the different interfaces (FIWARE, FILE, MQTT)
        Takes the data from the DataTransferModel and prepares the data for the output
        (Creates a OutputDataModel for the use in Function `send_outputs()`).

        Args:
            data_output (DataTransferModel): DataTransferModel with the output data
            from the calculation

        Returns:
            OutputDataModel: OutputDataModel with the output data as formatted data
        """
        output_data = OutputDataModel(entities=[])

        if data_output is None:
            logger.debug("No data for preparing the output.")
            return output_data
        output_attrs: dict[str, list] = {}
        output_cmds: dict[str, list] = {}

        if self.config is None:
            logger.error(
                "Configuration is not loaded. Please call prepare_start() first."
            )
            return output_data

        for component in data_output.components:
            for output in self.config.outputs:
                if output.id == component.entity_id:
                    self._process_attributes(component, output, output_attrs)
                    self._process_commands(component, output, output_cmds)

        for output in self.config.outputs:
            attributes = output_attrs.get(output.id, [])
            commands = output_cmds.get(output.id, [])
            output_data.entities.append(
                OutputDataEntityModel(
                    id=output.id, attributes=attributes, commands=commands
                )
            )

        return output_data

    def _process_attributes(
        self,
        component: DataTransferComponentModel,
        output: OutputModel,
        output_attrs: dict,
    ) -> dict:
        """Helper function to process attributes."""
        for attribute in output.attributes:
            if attribute.id == component.attribute_id:
                if output.id not in output_attrs:
                    output_attrs[output.id] = []

                attribute.value = (
                    component.value if component.value is not None else attribute.value
                )

                output_attrs[output.id].append(
                    AttributeModel(
                        id=attribute.id,
                        value=attribute.value,
                        unit=component.unit,
                        timestamp=component.timestamp,
                    )
                )
                break
        return output_attrs

    def _process_commands(
        self,
        component: DataTransferComponentModel,
        output: OutputModel,
        output_cmds: dict,
    ) -> dict:
        """Helper function to process commands."""
        for command in output.commands:
            if command.id == component.attribute_id:
                if output.id not in output_cmds:
                    output_cmds[output.id] = []

                # TODO: type checking necessary? Dataframes and bools not allowed for commands
                command.value = (
                    component.value if component.value is not None else command.value
                )

                output_cmds[output.id].append(
                    CommandModel(id=command.id, value=command.value)
                )
                break
        return output_cmds

    def cleanup_service(self):
        """
        Cleanup the service resources:
            - MQTT Client
        If more resources are added in the future, make sure to clean them up here.
        """

        self.stop_mqtt_client()
        logger.debug("Service stopped, cleanup finished.")

    async def start_service(self):
        """
        Main function for converting the data
        """

        sampling_time = (
            self.config.controller_settings.time_settings.calculation.sampling_time
            * get_time_unit_seconds(
                self.config.controller_settings.time_settings.calculation.sampling_time_unit
            )
        )

        logger.info("Start the Service")
        # Hold the sampling time at the beginning,
        # so that the mqtt interfaces is ready and data is available
        if self.config.interfaces.mqtt:
            await self._hold_sampling_time(
                start_time=datetime.now(), hold_time=sampling_time
            )

        while not self.shutdown_event.is_set():
            logger.debug("Start the Prozess")
            start_time = datetime.now()

            if self.config.interfaces.fiware:
                self.update_authentication()

            data_input = await self.get_data(method=DataQueryTypes.CALCULATION)

            if data_input is not None:
                data_output = await self.calculation(data=data_input)

                data_output = self.prepare_output(data_output=data_output)

                await self.send_outputs(data_output=data_output)

            await self._set_health_timestamp()

            await self._hold_sampling_time(
                start_time=start_time, hold_time=sampling_time
            )

        logger.debug("Service will be stopped, running cleanup")
        self.cleanup_service()

    async def start_calibration(self):
        """
        Function for autonomous adjustment of the system parameters
        """

        if self.config.controller_settings.time_settings.calibration is None:
            logger.info(
                "No Information about the calibration time in the configuration. "
                "Calibration will not be performed."
            )
            return

        sampling_time = (
            self.config.controller_settings.time_settings.calibration.sampling_time
            * get_time_unit_seconds(
                self.config.controller_settings.time_settings.calibration.sampling_time_unit
            )
        )

        while not self.shutdown_event.is_set():
            logger.debug("Start Calibration")
            start_time = datetime.now()
            data_input = await self.get_data(method=DataQueryTypes.CALIBRATION)
            await self.calibration(data=data_input)

            await self._hold_sampling_time(
                start_time=start_time, hold_time=sampling_time
            )
        logger.debug("Calibration was stopped")

    async def check_health_status(self):
        """
        Function to check the health-status of the service
        """
        logger.debug("Start the the Health-Check")
        while not self.shutdown_event.is_set():
            start_time = datetime.now()
            sampling_time = (
                self.config.controller_settings.time_settings.calculation.sampling_time
                * get_time_unit_seconds(
                    self.config.controller_settings.time_settings.calculation.sampling_time_unit
                )
            )

            await update_health_file(
                time_cycle=sampling_time,
                timestamp_health=self.timestamp_health,
                timestamp_now=start_time,
            )

            await self._hold_sampling_time(start_time=start_time, hold_time=10)

    async def _set_health_timestamp(self):
        """
        Function to set the timestamp of the last health-check
        """
        self.timestamp_health = datetime.now()
        return



"""
Description: This file contains the models for the configuration of the system controller.
Authors: Martin Altenburger
"""

import json
from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from pydantic import BaseModel, ConfigDict, ValidationError, Field
from pydantic.functional_validators import model_validator
from loguru import logger
from pandas import DataFrame
from filip.models.base import DataType
from encodapy.config.types import (
    AttributeTypes,
    Interfaces,
    TimerangeTypes,
)
from encodapy.utils.error_handling import ConfigError, InterfaceNotActive
from encodapy.utils.units import DataUnits, TimeUnits
from encodapy.components.basic_component_config import ControllerComponentModel


class InterfaceModel(BaseModel):
    """Base class for the interfaces
    TODO: - How to use this model?
    """

    mqtt: bool = False
    fiware: bool = False
    file: bool = False


class AttributeModel(BaseModel):
    """
    Base class for the attributes

    Contains:
    - id: The id of the attribute
    - id_interface: The id of the attribute on the interface (if not set, the id is used)
    - type: The type of the attribute
    - value: The value of the attribute
    - unit: The unit of the attribute
    - datatype: The datatype of the attribute
    - timestamp: The timestamp of the attribute

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: str
    id_interface: str = Field(default=None)
    type: AttributeTypes = AttributeTypes.VALUE
    value: Union[str, float, int, bool, Dict, List, DataFrame, None] = None
    unit: Union[DataUnits, None] = None
    datatype: DataType = DataType("Number")
    timestamp: Union[datetime, None] = None

    @model_validator(mode="after")
    def set_id_interface(self) -> "AttributeModel":
        """
        Sets the 'id_interface' attribute to the value of 'id' if it is currently None.

        Returns:
            AttributeModel: The instance with the updated 'id_interface' attribute.
        """
        if self.id_interface is None:
            self.id_interface = self.id
        return self


class CommandModel(BaseModel):
    """
    Base class for the commands

    Contains:
    - id: The id of the command
    - id_interface: The id of the command on the interface (if not set, the id is used)
    - value: The value of the command
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: str
    id_interface: str = Field(default=None)
    value: Union[str, int, float, List, Dict, None] = None

    @model_validator(mode="after")
    def set_id_interface(self) -> "CommandModel":
        """Sets the 'id_interface' attribute to the value of 'id'
        if it is currently None.
        """
        if self.id_interface is None:
            self.id_interface = self.id
        return self


class InputModel(BaseModel):
    """
    Model for the configuration of inputs.

    Contains:
    - id: The id of the input
    - interface: The interface of the input
    - id_interface: The id of the input on the interface, default is the id
    - attributes: A list of attributes for the input as AttributeModel

    """

    id: str
    interface: Interfaces
    id_interface: str = Field(default=None)
    attributes: list[AttributeModel]

    @model_validator(mode="after")
    def set_id_interface(self) -> "InputModel":
        """Sets the 'id_interface' attribute to the value of 'id'
        if it is currently None.
        """
        if self.id_interface is None:
            self.id_interface = self.id
        return self


class StaticDataModel(InputModel):
    """
    Model for the configuration of inputs.

    Contains:
    - id: The id of the input
    - interface: The interface of the input
    - id_interface: The id of the input on the interface
    - attributes: A list of attributes for the static data as AttributeModel

    """


class OutputModel(BaseModel):
    """
    Model for the configuration of outputs.

    Contains:
    - id: The id of the output
    - interface: The interface of the output
    - id_interface: The id of the output on the interface, default is the id
    - attributes: The attributes of the output
    """

    id: str
    interface: Interfaces
    id_interface: str = Field(default=None)
    attributes: list[AttributeModel]
    commands: list[CommandModel]

    @model_validator(mode="after")
    def set_id_interface(self) -> "OutputModel":
        """Sets the 'id_interface' attribute to the value of 'id'
        if it is currently None.
        """
        if self.id_interface is None:
            self.id_interface = self.id
        return self


class TimeSettingsCalculationModel(BaseModel):
    """
    Base class for the calculation time settings of the controller / system.

    Contains:
    - timerange: The timerange for the calculation (if only one value is needed and primary value,
    otherwise use timerange_min and timerange_max)
    - timerange_min: The minimum timerange for the calculation (only used if timerange is not set
    and timerange_max is set too)
    - timerange_max: The maximum timerange for the calculation (only used if timerange is not set
    and timerange_min is set too)
    - timerange_type: Type of time period, relative to the last result or absolute at the current
    time (if not set, the default type is absolute)
    - timerange_unit: The unit of the timerange (if not set, the default unit is minute)
    - timestep: The timestep for the calculation (if not set, the default value is 1), the
    related unit is defined in the timestep_unit attribute
    - timestep_unit: The unit of the timestep (if not set, the default unit is second)
    - sampling_time: The sampling time for the calculation (if not set, the default value is 1),
    the related unit is defined in the sampling_time_unit attribute
    - sampling_time_unit: The unit of the sampling time (if not set, the default unit is minute)
    """

    timerange: Optional[float] = None
    timerange_min: Optional[float] = None
    timerange_max: Optional[float] = None
    timerange_type: Optional[TimerangeTypes] = TimerangeTypes.ABSOLUTE
    timerange_unit: Optional[TimeUnits] = TimeUnits.MINUTE

    timestep: Union[float, int] = 1
    timestep_unit: TimeUnits = TimeUnits.SECOND

    sampling_time: Union[float, int] = 1
    sampling_time_unit: TimeUnits = TimeUnits.MINUTE

    @model_validator(mode="after")
    def check_timerange_parameters(self) -> "TimeSettingsCalculationModel":
        """Check the timerange parameters.

        Raises:
            ValueError: If the timerange parameters are not set correctly

        Returns:
            TimeSettingsCalculationModel: The model with the validated parameters
        """

        if self.timerange is None and (
            self.timerange_min is None or self.timerange_max is None
        ):
            raise ValueError(
                "Either 'timerange' or 'timerange_min' and 'timerange_max' must be set."
            )

        if self.timerange is not None and (
            self.timerange_min is not None or self.timerange_max is not None
        ):
            logger.warning(
                "Either 'timerange' or both 'timerange_min' and 'timerange_max' should be set, \
                not both. Using 'timerange' as the only value."
            )

            self.timerange_min = None
            self.timerange_max = None

        return self

    @model_validator(mode="before")
    @classmethod
    def check_timestep_units(cls, data: Any) -> Any:
        """Checks the units of the times in the configuration and provides feedback (debug logging)
        if the default values are used

        Args:
            data (Any): The data to check (input data)

        Returns:
            Any: The data with the checked timestep units
        """
        if "timestep" not in data:
            logger.debug("No timestep is set - using default value '1'")
        if "timestep_unit" not in data:
            logger.debug("No timestep unit is set - using default unit 'second'")
        if "sampling_time" not in data:
            logger.debug("No sampling time is set - using default value '1'")
        if "sampling_time_unit" not in data:
            logger.debug("No sampling time unit is set - using default unit 'minute'")
        if "timerange_unit" not in data:
            logger.debug("No timerange unit is set - using default unit 'minute'")
        if "timerange_type" not in data:
            logger.debug("No timerange type is set - using default type 'absolute'")
        return data


class TimeSettingsCalibrationModel(BaseModel):
    """
    Base class for the calibration time settings of the controller / system.

    Contains:
    - timerange: The timerange for the calibration (if not set, the default value is 1),
    the related unit is defined in the timerange_unit attribute
    - timerange_unit: The unit of the timerange (if not set, the default unit is minute)
    - timestep: The timestep for the calibration (if not set, the default value is 1),
    the related unit is defined in the timestep_unit attribute
    - timestep_unit: The unit of the timestep (if not set, the default unit is second)
    - sampling_time: The sampling time for the calibration (if not set, the default value is 1),
    the related unit is defined in the sampling_time_unit attribute
    - sampling_time_unit: The unit of the sampling time (if not set, the default unit is day)

    """

    timerange: Optional[float] = None
    timerange_unit: Optional[TimeUnits] = TimeUnits.MINUTE

    timestep: Union[float, int] = 1
    timestep_unit: TimeUnits = TimeUnits.SECOND

    sampling_time: Union[float, int] = 1
    sampling_time_unit: TimeUnits = TimeUnits.DAY


class TimeSettingsResultsModel(BaseModel):
    """
    Settings for the timesteps of the results.
    """

    timestep: Union[float, int] = 1
    timestep_unit: TimeUnits = TimeUnits.SECOND


class TimeSettingsModel(BaseModel):
    """
    Base class for the time settings of the controller / system.

    Contains:
    - calculation: The timeranges and settings für the calculation
    - calibration: The timeranges and settings for the calibration
    - results: The timesettings for the results

    TODO: Add the needed fields - calibration?
    """

    calculation: TimeSettingsCalculationModel
    calibration: Optional[TimeSettingsCalibrationModel] = None
    results: Optional[TimeSettingsResultsModel] = None


class ControllerSettingModel(BaseModel):
    """
    Model for the configuration of the controller settings.

    Contains:
    - time_settings: The time settings for the controller
    - specific_settings: The specific settings for the controller - not defined as a model

    TODO: What is needed here?
    """

    time_settings: TimeSettingsModel
    specific_settings: Optional[dict] = {}


class ConfigModel(BaseModel):
    """
    Base Model for the configuration

    Contains:
    - interfaces: The interfaces of the system controller
    - inputs: The inputs of the system controller
    - outputs: The outputs of the system controller
    - staticdata: The static configuration data for devices the system controller
    - controller_components: The components of the controller
    - controller_settings: The settings for the controller
    """

    interfaces: InterfaceModel

    inputs: list[InputModel]
    outputs: list[OutputModel]
    staticdata: list[StaticDataModel]

    controller_components: list[ControllerComponentModel]

    controller_settings: ControllerSettingModel

    @classmethod
    def from_json(cls, file_path: str):
        """
        Load the configuration from a JSON file.
        """

        try:
            with open(file_path, "r", encoding="utf-8") as f:
                config_data = json.load(f)
            return cls(**config_data)
        except (FileNotFoundError, json.JSONDecodeError) as e:
            logger.error(f"Couldn't load the json file: {e}")
        except ValidationError as e:
            logger.error(e)

        raise ConfigError("Coudn't load configuration from json file")

    @model_validator(mode="after")
    def check_interfaces(self) -> "ConfigModel":
        """Check the interfaces of the input / output.

        Raises:
            ValueError: If no interface is set or the interface is not active \
                and in use for input or output

        Returns:
            ConfigModel: The model with the validated interfaces
        """

        if not any(
            [self.interfaces.mqtt, self.interfaces.fiware, self.interfaces.file]
        ):
            raise InterfaceNotActive("At least one interface must be set.")

        def check_interface_active(
            datapoints: Union[
                List[InputModel], List[StaticDataModel], List[OutputModel]
            ],
        ) -> None:
            """
            Check if the interface is active for the datapoints.

            Args:
                datapoints (Union[InputModel, StaticDataModel, OutputModel]): \
                    The datapoints to check

            Raises:
                ValueError: If the datapoints are not valid
                InterfaceNotActive: If the interface is not active \
                    but used for the datapoints
            """
            if len(datapoints) == 0:
                return
            if isinstance(datapoints[0], InputModel):
                config_part = "input"
            elif isinstance(datapoints[0], StaticDataModel):
                config_part = "static data"
            elif isinstance(datapoints[0], OutputModel):
                config_part = "output"
            else:
                raise ValueError("The input data is not valid")

            for datapoint in datapoints:
                if datapoint.interface == Interfaces.MQTT and not self.interfaces.mqtt:
                    raise InterfaceNotActive(
                        f"The MQTT interface is used for the {config_part} '{datapoint.id}', "
                        "but not set in the configuration."
                    )
                if (
                    datapoint.interface == Interfaces.FIWARE
                    and not self.interfaces.fiware
                ):
                    raise InterfaceNotActive(
                        f"The FIWARE interface is used for the {config_part} '{datapoint.id}', "
                        "but not set in the configuration."
                    )
                if datapoint.interface == Interfaces.FILE and not self.interfaces.file:
                    raise InterfaceNotActive(
                        f"The FILE interface is used for the {config_part} '{datapoint.id}', "
                        "but not set in the configuration."
                    )

        check_interface_active(self.inputs)
        check_interface_active(self.staticdata)
        check_interface_active(self.outputs)

        return self


class StaticDataFileAttribute(BaseModel):
    """
    Model for static data file attributes.
    
    Contains:
        id (str): The unique identifier for the attribute.
        value (Union[str, float, int, bool, Dict, List, DataFrame, None]): \
            The value of the attribute.
        metadata (Union[dict[str, str], None]): Metadata dictionary or None.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: str
    value: Union[str, float, int, bool, Dict, List, DataFrame, None]
    metadata: Optional[Dict] = None


class StaticDataFileEntity(BaseModel):
    """
    Model for static data file entities.

    Contains:
        id (str): The unique identifier for the entity.
        attributes (list[StaticDataFileAttribute]): The attributes of the entity.
    """

    id: str
    attributes: list[StaticDataFileAttribute]


class StaticDataFile(BaseModel):
    """
    Model for static data files.

    Contains:
        staticdata (list[StaticDataFileEntity]): The static data entities.
    """

    staticdata: list[StaticDataFileEntity]



"""
Description: This file contains the class MqttConnection,
which is used to store the connection parameters for the MQTT broker.
Author: Maximilian Beyer
"""

import json
import os
import re
from datetime import datetime, timezone
from typing import Optional, Union

import paho.mqtt.client as mqtt
from loguru import logger
from paho.mqtt.enums import CallbackAPIVersion
from pandas import DataFrame

from encodapy.config import (
    ConfigModel,
    DataQueryTypes,
    DefaultEnvVariables,
    InputModel,
    Interfaces,
    OutputModel,
)
from encodapy.utils.error_handling import ConfigError, NotSupportedError
from encodapy.utils.models import (
    AttributeModel,
    InputDataAttributeModel,
    InputDataEntityModel,
    OutputDataEntityModel,
)


class MqttConnection:
    """
    Class for the connection to a MQTT broker.
    Only a helper class.
    """

    def __init__(self) -> None:
        """
        Constructor for the MqttConnection class.
        Initializes the MQTT parameters and the MQTT client.
        """
        self.mqtt_params: dict = {}
        self.config: ConfigModel
        self.mqtt_client: Optional[mqtt.Client] = None
        self.mqtt_message_store: dict[str, dict] = {}
        self._mqtt_loop_running = False

    def load_mqtt_params(self) -> None:
        """
        Function to load the MQTT parameters from the environment variables
        or use the default values from the DefaultEnvVariables class.
        """
        # the IP of the broker
        self.mqtt_params["host"] = os.environ.get(
            "MQTT_HOST", DefaultEnvVariables.MQTT_HOST.value
        )
        # the port of the broker
        self.mqtt_params["port"] = int(
            os.environ.get("MQTT_PORT", DefaultEnvVariables.MQTT_PORT.value)
        )
        # the username to connect to the broker
        self.mqtt_params["username"] = os.environ.get(
            "MQTT_USERNAME", DefaultEnvVariables.MQTT_USERNAME.value
        )
        # the password to connect to the broker
        self.mqtt_params["password"] = os.environ.get(
            "MQTT_PASSWORD", DefaultEnvVariables.MQTT_PASSWORD.value
        )
        # the topic prefix to use for the topics
        self.mqtt_params["topic_prefix"] = os.environ.get(
            "MQTT_TOPIC_PREFIX", DefaultEnvVariables.MQTT_TOPIC_PREFIX.value
        )

        if not self.mqtt_params["host"] or not self.mqtt_params["port"]:
            raise ConfigError("MQTT host and port must be set")

    def prepare_mqtt_connection(self) -> None:
        """
        Function to prepare the MQTT connection
        """
        # initialize the MQTT client
        if not self.mqtt_client:
            self.mqtt_client = mqtt.Client(
                callback_api_version=CallbackAPIVersion.VERSION2
            )

        # set username and password for the MQTT client
        self.mqtt_client.username_pw_set(
            username=self.mqtt_params["username"], password=self.mqtt_params["password"]
        )

        # try to connect to the MQTT broker
        try:
            self.mqtt_client.connect(
                host=self.mqtt_params["host"], port=self.mqtt_params["port"]
            )
        except Exception as e:
            raise ConfigError(
                f"Could not connect to MQTT broker {self.mqtt_params['host']}:"
                f"{self.mqtt_params['port']} with given login information - {e}"
            ) from e

        # prepare the message store
        self.prepare_mqtt_message_store()

        # subscribe to all topics in the message store
        self.subscribe_to_message_store_topics()

        # start the MQTT client loop
        self.start_mqtt_client()

    def prepare_mqtt_message_store(self) -> None:
        """
        Function to prepare the MQTT message store for all input entities and their attributes.
        Sets the optional default values for all attributes of the entities in the config.
        The Topic structure is defined as follows:
        <topic_prefix>/<entity_id(_interface)>/<attribute_id(_interface)>

        Format of the message store:
        {
            "topic": {
                "entity_id": "entity_id",
                "attribute_id": "attribute_id",
                "payload": value,
                "timestamp": datetime.now(),
            }
        }
        """
        if self.mqtt_message_store:
            logger.warning("MQTT message store is not empty and will be overwritten.")
            self.mqtt_message_store.clear()

        if self.config is None:
            raise ConfigError(
                "ConfigModel is not set. Please set the config before using the MQTT connection."
            )

        for entity in self.config.inputs:
            if entity.interface == Interfaces.MQTT:
                # add the entity itself to the message store
                topic = self.assemble_topic_parts(
                    [self.mqtt_params["topic_prefix"], entity.id_interface]
                )

                self._add_item_to_mqtt_message_store(
                    topic=topic,
                    entity_id=entity.id,
                )

                # iterate over all attributes of the entity and add them to the message store
                for attribute in entity.attributes:
                    topic = self.assemble_topic_parts(
                        [
                            self.mqtt_params["topic_prefix"],
                            entity.id_interface,
                            attribute.id_interface,
                        ]
                    )

                    # set the default value for the attribute
                    if hasattr(attribute, "value"):
                        default_value = attribute.value
                    else:
                        default_value = None

                    self._add_item_to_mqtt_message_store(
                        topic=topic,
                        entity_id=entity.id,
                        attribute_id=attribute.id,
                        payload=default_value,
                    )

    def assemble_topic_parts(self, parts: list[str | None]) -> str:
        """
        Function to build a topic from a list of strings.
        Ensures that the resulting topic is correctly formatted with exactly one '/' between parts.

        Args:
            parts (list[str|None]): List of strings to be joined into a topic.

        Returns:
            str: The correctly formatted topic.

        Raises:
            ValueError: If the list of parts is empty.
        """
        if not parts:
            raise ValueError("The list of parts cannot be empty.")

        # drop a part if it is None or empty
        parts = [part for part in parts if part not in (None, "")]

        # Join the parts with a single '/',
        # stripping only trailing slashes from each part to avoid double slashes in the topic
        topic = "/".join(part.rstrip("/") for part in parts if isinstance(part, str))

        return topic

    def _add_item_to_mqtt_message_store(
        self,
        *,
        topic: str,
        entity_id: str,
        attribute_id: Optional[str] = None,
        payload=None,
        timestamp: Optional[datetime] = None,
    ) -> None:
        """
        Function to add an item to the MQTT message store.

        If the topic already exists, it will be overwritten and logs a warning.

        Args:
            topic (str): The topic to add the item to.
            entity_id (str): The ID of the entity.
            attribute_id (Optional[str]): The ID of the attribute (if applicable).
            payload: The default value of the message (if applicable).
            timestamp (Optional[datetime]): The initial timestamp of the message (if applicable).
        """
        if topic in self.mqtt_message_store:
            logger.warning(
                f"Topic {topic} from entity {entity_id} already exists in message store, "
                "overwriting it. This should not happen, check your configuration."
            )

        self.mqtt_message_store[topic] = {
            "entity_id": entity_id,
            "attribute_id": attribute_id,
            "payload": payload,
            "timestamp": timestamp,
        }

    def publish(
        self,
        topic: str,
        payload: Union[str, float, int, bool, dict, list, DataFrame, None],
    ) -> None:
        """
        Function to publish a message (payload) to a topic.

        Every payload is converted to a utf8 encoded string before publishing
        (at the latest from the paho-mqtt package used).

        Args:
            topic (str): The topic to publish the message to
            payload (Union[str, float, int, bool, dict, list, DataFrame, None]): payload to publish
        """
        if not self.mqtt_client:
            raise NotSupportedError(
                "MQTT client is not prepared. Call prepare_mqtt_connection() first."
            )

        payload = self.prepare_payload_for_publish(payload)
        self.mqtt_client.publish(topic, payload)
        logger.debug(f"Published to topic {topic}: {payload}")

    def prepare_payload_for_publish(
        self, payload: Union[str, float, int, bool, dict, list, DataFrame, None]
    ) -> Union[str, None]:
        """
        Function to prepare the payload for publishing.

        Converts the payload to a JSON string if it is a dict, list or DataFrame.
        If the payload is a string, float, int or bool, it is converted to a string.
        If the payload is None or an unsupported type, it is set to None.
        """
        try:
            if isinstance(payload, (dict, list)):
                payload = json.dumps(payload)
            elif isinstance(payload, DataFrame):
                payload = payload.to_json()
            elif isinstance(payload, (str, float, int, bool)):
                payload = str(payload)
            elif payload is None:
                pass
            else:
                logger.warning(
                    f"Unsupported payload type: {type(payload)}, set it to None"
                )
                payload = None

        except TypeError as e:
            logger.warning(f"Failed to convert payload: {e}, set it to None")
            payload = None

        return payload

    def subscribe_to_message_store_topics(self) -> None:
        """
        Function to subscribe to all topics in the message store.
        """
        if not self.mqtt_message_store:
            raise NotSupportedError(
                "MQTT message store is initialized, but empty. Cannot subscribe to topics."
            )

        for topic in self.mqtt_message_store:
            self.subscribe(topic)
            logger.debug(f"Subscribed to topic: {topic}")

    def subscribe(self, topic) -> None:
        """
        Function to subscribe to a topic
        """
        if not self.mqtt_client:
            raise NotSupportedError(
                "MQTT client is not prepared. Call prepare_mqtt_connection() first."
            )
        self.mqtt_client.subscribe(topic)

    def on_message(self, _, __, message):
        """
        Callback function for received messages.

        Stores the decoded message payload with its timestamp in the message store.
        If the message is from an entity, the payload is scanned for attributes and
        their values are stored in the message store, too.
        """
        if not hasattr(self, "mqtt_message_store"):
            raise NotSupportedError(
                "MQTT message store is not initialized. Call prepare_mqtt_connection() first."
            )

        current_time = datetime.now(timezone.utc)

        debug_message = (
            f"MQTT connection received message on {message.topic} at {current_time}."
        )

        if message.topic in self.mqtt_message_store:
            # decode the message payload
            try:
                payload = message.payload.decode("utf-8")
            except UnicodeDecodeError as e:
                logger.error(debug_message + f" Failed to decode message payload: {e}.")
                return

            # store payload and current time in the message store
            self.mqtt_message_store[message.topic]["payload"] = payload
            self.mqtt_message_store[message.topic]["timestamp"] = current_time

            debug_message += f" Updated MQTT message store with value: {payload}."

            # if the item in the store is from an entity, its attribute_id in the store must be None
            # and attribute values are possibly in payload
            if self.mqtt_message_store[message.topic]["attribute_id"] is None:
                # get the entity from the message store
                entity_id = self.mqtt_message_store[message.topic]["entity_id"]
                debug_message += (
                    f" Message is from entity {entity_id}, try to extract attributes."
                )
                # try to parse the payload as JSON
                try:
                    payload = json.loads(payload)
                    if isinstance(payload, dict):
                        # extract attributes from the payload and update the message store
                        logger.debug(debug_message)
                        self._extract_attributes_from_payload_and_update_store(
                            entity_id=entity_id, payload=payload, timestamp=current_time
                        )
                    else:
                        debug_message += (
                            f" Unexpected payload format, type={type(payload)}."
                        )
                        logger.warning(debug_message)
                except json.JSONDecodeError:
                    debug_message += (
                        f" Failed to decode JSON payload: type={type(payload)}."
                    )
                    logger.error(debug_message)
                    return
            else:
                logger.debug(debug_message)

    def _extract_attributes_from_payload_and_update_store(
        self,
        entity_id: str,
        payload: dict,
        timestamp: datetime = datetime.now(),
    ) -> None:
        """
        Function to extract attributes from the payload and update the message store.
        This is called when a message is received on a topic that corresponds to an entity.

        Args:
            entity_id (str): The ID of the entity the message is related to.
            payload (dict): The payload received from the MQTT broker.
            timestamp (datetime): The timestamp when the message was received.
        """
        if not hasattr(self, "mqtt_message_store"):
            raise NotSupportedError(
                "MQTT message store is not initialized. Call prepare_mqtt_connection() first."
            )

        debug_message = ""

        for key, value in payload.items():
            # search in the message store for a subtopic that matches the key and entity
            for topic, item in self.mqtt_message_store.items():
                # check if the item in the message store is from the entity
                if item["entity_id"] != entity_id:
                    continue

                # get subtopic (last part of the topic),
                # which could reference the attribute.id_interface
                subtopic = topic.split("/")[-1] if "/" in topic else topic

                # if subtopic matches key in the payload from entity, update the message store
                if subtopic == key:
                    item["payload"] = value
                    item["timestamp"] = timestamp
                    debug_message += (
                        f"Updated MQTT message store for topic {topic} with value: {value} "
                        f"and timestamp: {timestamp}"
                    )
                    continue

        if debug_message == "":
            debug_message += (
                f" No updates made to MQTT message store for entity {entity_id}."
            )

        logger.debug(debug_message)

    def start_mqtt_client(self):
        """
        Function to hang in on_message hook and start the MQTT client loop
        """
        if not hasattr(self, "mqtt_client") or self.mqtt_client is None:
            raise NotSupportedError(
                "MQTT client is not prepared. Call prepare_mqtt_connection() first."
            )

        if hasattr(self, "_mqtt_loop_running") and self._mqtt_loop_running:
            raise NotSupportedError("MQTT client loop is already running.")

        self.mqtt_client.on_message = self.on_message
        self.mqtt_client.loop_start()
        self._mqtt_loop_running = True  # state variable to check if the loop is running

    def stop_mqtt_client(self):
        """
        Function to stop the MQTT client loop and clean up resources
        """
        if isinstance(self.mqtt_client, mqtt.Client) and self._mqtt_loop_running:
            self.mqtt_client.loop_stop()
            self.mqtt_client.disconnect()
            self._mqtt_loop_running = False

    def get_data_from_mqtt(
        self,
        method: DataQueryTypes,  # pylint: disable=unused-argument
        entity: InputModel,
    ) -> InputDataEntityModel:
        """
        Function to get the data from the MQTT broker.
        It checks the MQTT message store for the topics of the entity and its attributes.
        If the topic is found, it extracts the data from the message payload.
        If the topic is not found or the payload is not in the expected format,
        it sets the data to None and marks it as unavailable.


        Args:
            method (DataQueryTypes): The method is currently not used.
            entity (InputModel): Input entity

        Returns:
            InputDataEntityModel: Model with input data (data=None if no data available)
        """
        if not hasattr(self, "mqtt_message_store"):
            raise NotSupportedError(
                "MQTT message store is not initialized. Call prepare_mqtt_connection() first."
            )

        attributes_values = []

        for attribute in entity.attributes:
            # construct the topic for the attribute
            topic = self.assemble_topic_parts(
                [
                    self.mqtt_params["topic_prefix"],
                    entity.id_interface,
                    attribute.id_interface,
                ]
            )

            # if the topic is not in the message store, mark the data as unavailable
            if topic not in self.mqtt_message_store:
                logger.warning(
                    f"Topic {topic} not found in MQTT message store. Setting data as None and "
                    "unavailable. User should check for possible misconfiguration!"
                )
                data = None
                data_available = False
                timestamp = None

            # if the topic is in the message store, extract the data from message payload
            else:
                message_payload = self.mqtt_message_store[topic]["payload"]
                try:
                    data = self._extract_payload_value(message_payload)
                    data_available = True
                except ValueError as e:
                    logger.error(
                        f"Failed to extract payload value for topic {topic}: {e}. "
                        "Setting data as None and unavailable."
                    )
                    data = None
                    data_available = False

                # Get the timestamp from the message store
                timestamp = self.mqtt_message_store[topic]["timestamp"]

            attributes_values.append(
                InputDataAttributeModel(
                    id=attribute.id,
                    data=data,
                    data_type=attribute.type,
                    data_available=data_available,
                    latest_timestamp_input=timestamp,
                    unit=None,  # TODO MB: Add unit handling if necessary
                )
            )
        return InputDataEntityModel(id=entity.id, attributes=attributes_values)

    def _extract_payload_value(
        self, payload
    ) -> Union[str, float, int, bool, dict, list, DataFrame, None]:
        """
        Function to extract data from the payload as needed.
        # TODO MB: How to use pd.read_json here for Dataframes?
        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_json.html
        """
        if payload is None or payload == "":
            return None

        # If the payload is not a string (maybe from other source), return it directly
        if not isinstance(payload, str):
            return payload

        # Try to parse JSON (automatically handles int, float, bool, dicts, lists)
        try:
            parsed = json.loads(payload)
            # If the payload is a valid dict, try to extract a value from it
            if isinstance(parsed, dict):
                # Ensure case-insensitive key check and return value of first found "value" key
                value = next((parsed[k] for k in parsed if k.lower() == "value"), None)
                if value is not None:
                    return value
            return parsed
        except json.JSONDecodeError:
            pass

        # If the payload is a string that starts with a number, try to extract it
        # This handles cases like "22.5 °C" or "6552.0 h" where we want to extract the number
        # and ignore the unit (if any).
        # The regex matches an optional leading '-' for negative numbers, followed by digits,
        # optionally with a decimal point and more digits.
        match = re.match(r"^\s*(-?\d+(\.\d+)?)", payload)
        if match:
            num_str = match.group(1)
            if "." in num_str:
                return float(num_str)
            return int(num_str)

        # if nothing else worked, return the payload as is
        return payload

    def send_data_to_mqtt(
        self,
        output_entity: OutputModel,
        output_attributes: list[AttributeModel],
        # output_commands: list[CommandModel],
    ) -> None:
        """
        Function to send the output data to MQTT (publish the data to the MQTT broker).

        Args:
            - output_entity: OutputModel with the output entity
            - output_attributes: list with the output attributes
        """
        if not hasattr(self, "mqtt_client"):
            raise NotSupportedError(
                "MQTT client is not prepared. Call prepare_mqtt_connection() first."
            )

        # check if the config is set
        if self.config is None:
            raise ConfigError(
                "ConfigModel is not set. Please set the config before using the MQTT connection."
            )

        # publish the data to the MQTT broker
        for attribute in output_attributes:
            topic = self.assemble_topic_parts(
                [
                    self.mqtt_params["topic_prefix"],
                    output_entity.id_interface,
                    attribute.id_interface,
                ]
            )
            payload = attribute.value
            self.publish(topic, payload)

    def _get_last_timestamp_for_mqtt_output(
        self, output_entity: OutputModel
    ) -> tuple[OutputDataEntityModel, Union[datetime, None]]:
        """
        Function to get the latest timestamps of the output entity from a MQTT message, if exists.

        Args:
            output_entity (OutputModel): Output entity

        Returns:
            tuple[OutputDataEntityModel, Union[datetime, None]]:
                - OutputDataEntityModel with timestamps for the attributes
                - the latest timestamp of the output entity for the attribute
                with the oldest value (None if no timestamp is available)
        """
        timestamps: list = []
        timestamp_latest_output = None

        return (
            OutputDataEntityModel(id=output_entity.id, attributes_status=timestamps),
            timestamp_latest_output,
        )



"""
Description: This module is used to update the health file for the health check.\
    The health file is checked by the healthcheck.
Author: Martin Altenburger
"""

from datetime import datetime, timedelta
from typing import Union
from loguru import logger


async def update_health_file(
    time_cycle: int, timestamp_health: Union[datetime, None], timestamp_now: datetime
):
    """
    create a file that is checked by healthcheck

    Args:
        - time_cycle: int, time in minutes for the health check
        - timestamp_health: datetime, timestamp of the last health check or None,\
            if no health check was done
        - timestamp_now: datetime, current timestamp
    """
    if timestamp_health is None:
        timestamp_health = timestamp_now
        with open("health", "w", encoding="utf-8") as f:
            f.write("OK")
    elif timestamp_health >= (timestamp_now - timedelta(minutes=time_cycle)):
        with open("health", "w", encoding="utf-8") as f:
            f.write("OK")
    else:
        logger.debug("Health-Status not ok - skip writing health file")
        return

    return



"""
Description: LoggerControl class to control the log level of the application.
Authors: Martin Altenburger
"""

import os
import sys
from loguru import logger
from encodapy.config.default_values import DefaultEnvVariables


class LoggerControl:
    """
    LoggerControl class for the control of the log level of the application.
    """

    def __init__(self) -> None:

        log_level = os.environ.get(
            "LOG_LEVEL", DefaultEnvVariables.LOG_LEVEL.value
        ).upper()

        logger.remove()
        logger.add(sys.stdout, level=log_level)



"""
Description: Enum classes for the types in the configuration
Authors: Martin Altenburger
"""

from enum import Enum


class Interfaces(Enum):
    """
    Enum class for the interfaces
    Possible values:
    - MQTT (mqtt): MQTT interface
    - FIWARE (fiware): FIWARE interface
    - FILE (file): File interface
    """

    MQTT = "mqtt"
    FIWARE = "fiware"
    FILE = "file"


class AttributeTypes(Enum):
    """
    Enum class for the attribute types
    Possible values:
    - TIMESERIES (timeseries): Timeseries data
    - VALUE (value): Single value data
    """

    TIMESERIES = "timeseries"
    VALUE = "value"


class TimerangeTypes(Enum):
    """Enum class for the timedelta types

    Possible values:
    - ABSOLUTE (absolute): The timedelta is calculated from the actual time
    - RELATIVE (relative): The timedelta is calculated from the last timestamp
    """

    ABSOLUTE = "absolute"
    RELATIVE = "relative"


class DataQueryTypes(Enum):
    """
    Enum class for the data query types
    Possible values:
    - CALCULATION (calculation): Calculation of the data
    - CALIBRATION (calibration): Calibration of the data
    """

    CALCULATION = "calculation"
    CALIBRATION = "calibration"


class FileExtensionTypes(Enum):
    """
    Enum class for file Extensions
    Possible values:
    - CSV (csv): Comma-separated values
    - JSON (json): JavaScript Object Notation
    """

    CSV = ".csv"
    JSON = ".json"



"""
This package contains modules for handling different types of connections
for the TWE-Flex project. It includes connections for FIWARE and file-based
communication.

Modules:
    fiware_connection: Contains the FiwareConnection class for handling FIWARE communication.
    file_connection: Contains the FileConnection class for handling file-based communication.
"""

from .fiware_connection import FiwareConnection
from .file_connection import FileConnection
from .mqtt_connection import MqttConnection



"""
Description: This file contains the error handling classes for the system controller.
Author: Martin Altenburger
"""


class ConfigError(Exception):
    """Exception raised for errors in the configuration file.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class NotSupportedError(Exception):
    """Exception raised for a configuration which is not suported yet.

    Attributes:
        message -- explanation of the error
    """

    def __init__(
        self,
        message: str = "The configuration is not supported yet.\
            Please contact the developer for further information.",
    ):
        self.message = message
        super().__init__(self.message)


class NoCredentials(Exception):
    """Exception raised if there are no / not all credentials given in the env.

    Attributes:
        message -- explanation of the error
    """

    def __init__(
        self,
        message: str = "Not all authentication credentials are available",
    ):
        self.message = message
        super().__init__(self.message)


class InterfaceNotActive(Exception):
    """Exception raised if the interface is not active.

    Attributes:
        message -- explanation of the error
    """

    def __init__(
        self,
        message: str = "The interface is not active, could not get the data",
    ):
        self.message = message
        super().__init__(self.message)



"""
Description: This file contains the models for the use in the system controller itself.
Authors: Martin Altenburger
"""

from datetime import datetime
from typing import Dict, List, Optional, Union
from pandas import DataFrame
from pydantic import BaseModel, ConfigDict
from filip.models.ngsi_v2.base import NamedMetadata
from filip.models.ngsi_v2.context import ContextEntity
from encodapy.config.models import AttributeModel, CommandModel
from encodapy.config.types import AttributeTypes
from encodapy.utils.units import DataUnits


class InputDataAttributeModel(BaseModel):
    """
    Model for a attribute of input data of the system controller.

    Contains:
    - id: The id of the input data attribute
    - data: The input data as a DataFrame or a single value
    - unit: The unit of the input data
    - data_type: The type of the input data (AttributeTypes)
    - data_available: If the data is available
    - latest_timestamp_input: The latest timestamp of the input data from the query or None,\
        if the data is not available
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: str
    data: Union[str, float, int, bool, Dict, List, DataFrame, None]
    unit: Union[DataUnits, None] = None
    data_type: AttributeTypes
    data_available: bool
    latest_timestamp_input: Union[datetime, None]


class InputDataEntityModel(BaseModel):
    """
    Model for the input data of the system controller.

    Contains:
    - id: The id of the input data entity
    - attributes: List of the input data attributes as InputDataAttributeModel
    """

    id: str
    attributes: List[InputDataAttributeModel]


class StaticDataEntityModel(InputDataEntityModel):
    """
    Model for the static data of the system controller, same like InputDataEntityModel.

    Contains:
    - id: The id of the input data entity
    - attributes: List of the input data attributes as InputDataAttributeModel
    """


class OutputDataAttributeModel(BaseModel):
    """
    Model for a attribute of output data of the system controller - status based on the status\
        of the interface.

    Contains:
    - id: The id of the output data attribute
    - latest_timestamp_output: The latest timestamp of the output data from the query or None,\
        if the data is not available
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: str
    latest_timestamp_output: Optional[Union[datetime, None]] = None


class OutputDataEntityModel(BaseModel):
    """
    Model for the status of the output data of the system controller.

    Contains:
    - id: The id of the output entity
    - latest_timestamp_output: The latest timestamp of the output data from the query or None,\
        if the data is not available

    - attributes: List of the output data attributes as OutputDataAttributeModel
    - commands: List of the output data commands as OutputDataCommandModel

    """

    id: str
    attributes: Optional[List[AttributeModel]] = []
    attributes_status: Optional[List[OutputDataAttributeModel]] = []
    commands: Optional[List[CommandModel]] = []


class InputDataModel(BaseModel):
    """
    Model for the input data of the system controller.

    Contains:
    - input_entities: List of the input data entities as InputDataEntityModel
    - output_entities: List of the output data entities as OutputDataEntityModel
    - static_entities: List of the static data entities as StaticDataEntityModel
    """

    input_entities: list[InputDataEntityModel]
    output_entities: list[OutputDataEntityModel]
    static_entities: list[StaticDataEntityModel]


class OutputDataModel(BaseModel):
    """
    Model for the output data of the system controller.

    Contains:
    - output_entitys: List of the output data entitys as OutputDataEntityModel
    """

    entities: list[OutputDataEntityModel]


class ComponentModel(BaseModel):
    """
    Model for the dataflow (input/output) of the controller.

    Contains:
    - entity: The entity (input / output) of the datapoint for the controller
    - attribute: The attribute of the datapoint for the controller
    """

    entity_id: str
    attribute_id: str


class DataTransferComponentModel(ComponentModel):
    """
    Model for the components of the data transfer between calculation and the basic service.

    Contains:
    - entity_id: The id of the entity of the component
    - attribute_id: The id of the attribute of the component
    - value: The output data value as OutputDataModel
    - unit: The unit of the output data
    - timestamp: The timestamp of the output
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    value: Union[str, float, int, bool, Dict, List, DataFrame, None]
    unit: Union[DataUnits, None] = None
    timestamp: Optional[Union[datetime, None]] = None


class DataTransferModel(BaseModel):
    """
    Model for the data transfer between calculation and the basic service.

    Contains:
    - components: List of the components of the data transfer as DataTransferComponentModel
    """

    components: list[DataTransferComponentModel] = []


# Models for the Fiware Connection
class MetaDataModel(BaseModel):
    """
    Model for the metadata of datapoints of the controller.

    Contains:
    - timestamp: The timestamp of the data
    - unit: The unit of the data
    """

    timestamp: Union[datetime, None] = None
    unit: Union[DataUnits, None] = None


class FiwareDatapointParameter(BaseModel):
    """
    Model for the Fiware datapoint parameter.
    Contains:
        entity (ContextEntity): The entity of the datapoint
        attribute (AttributeModel): The attribute of the datapoint
        metadata (list[NamedMetadata]): The metadata of the attribute
    Args:
        BaseModel (BaseModel): Pydantic BaseModel of a datapoint in fiware
    """

    entity: ContextEntity
    attribute: AttributeModel
    metadata: list[NamedMetadata]


class FiwareAuth(BaseModel):
    """
    Base model for the Fiware authentication.
    Contains:
        client_id (str): The client id
        client_secret (str): The client secret
        token_url (str): The token url
        baerer_token (str): The baerer token
    """

    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    token_url: Optional[str] = None
    baerer_token: Optional[str] = None


class FiwareParameter(BaseModel):
    """
    Model for the Fiware connection parameters.
    Contains:
        cb_url (str): The context broker url
        service (str): The service
        service_path (str): The service path
        authentication (Optional[Union[FiwareAuth, None]]): The authentication
    """

    cb_url: str
    service: str
    service_path: str
    authentication: Optional[Union[FiwareAuth, None]] = None


class DatabaseParameter(BaseModel):
    """
    Model for the database connection parameters.
    Contains:
        crate_db_url (str): The CrateDB url
        crate_db_user (Optional[str]): The CrateDB user
        crate_db_pw (Optional[str]): The CrateDB password
        crate_db_ssl (Optional[bool]): The CrateDB ssl
    """

    crate_db_url: str
    crate_db_user: Optional[Union[str, None]] = None
    crate_db_pw: Optional[str] = ""
    crate_db_ssl: Optional[bool] = True


class FiwareConnectionParameter(BaseModel):
    """
    Model for the Fiware connection parameters.
    Contains:
        fiware_params (FiwareParameter): The Fiware parameters
        database_params (DatabaseParameter): The database parameters
    """

    fiware_params: FiwareParameter
    database_params: DatabaseParameter



"""
Description: This module contains the definition of a service to run
    a component based on the configuration.
Author: Martin Altenburger
"""

from typing import Optional
import asyncio
from loguru import logger
from encodapy.service import ControllerBasicService
from encodapy.utils.models import (
    InputDataModel,
    InputDataEntityModel,
    InputDataAttributeModel,
    DataTransferModel,
    DataTransferComponentModel,
    AttributeTypes,
)
from encodapy.components.basic_component import BasicComponent
from encodapy.components.component_loader import get_component_class_model


class ComponentRunnerService(ControllerBasicService):
    """
    Class for a component runner service

    """

    def __init__(self, shutdown_event: Optional[asyncio.Event] = None) -> None:
        """
        Constructor for the ComponentRunnerService

        Args:
            shutdown_event (Optional[asyncio.Event]): \
                The shutdown event to signal when the service should stop
        """
        self.components: list[BasicComponent] = []
        super().__init__(shutdown_event=shutdown_event)

    def prepare_start(self):
        """ Function to prepare the service for start
        This function loads the configuration \
            and initializes the component.
        """
        for component in self.config.controller_components:

            if component.active is False:
                continue
            component_type = get_component_class_model(component.type)
            component = component_type(
                config=component, component_id=component.id, static_data=self.staticdata
            )
            self.components.append(component)

    def _result_to_input_data_attribute(
        self, result: DataTransferComponentModel
    ) -> InputDataAttributeModel:
        """
        Function to convert a DataTransferComponentModel to an InputDataAttributeModel
        """
        return InputDataAttributeModel(
            data_type=AttributeTypes.VALUE,
            id=result.attribute_id,
            data=result.value,
            unit=result.unit,
            latest_timestamp_input=result.timestamp,
            data_available=result.value is not None,
        )

    def _add_result_to_input_entity(
        self, result: DataTransferComponentModel, input_entity: InputDataEntityModel
    ) -> InputDataEntityModel:
        """
        Function to add the result of a component to an input entity, which exists.

        If the attribute already exists, it updates the attribute with the new value.
        If the attribute does not exist, it adds a new attribute to the entity.

        Args:
            result (DataTransferComponentModel): The result of the component to add
            input_entity (InputDataEntityModel): The input entity to add the result to
        Returns:
            InputDataEntityModel: The updated input entity
        """

        for attribute in input_entity.attributes:
            if result.attribute_id == attribute.id:
                attribute.data = result.value
                attribute.unit = result.unit
                logger.debug(
                    f"Update attribute: {attribute.id} with value: {result.value} "
                    f"and unit: {result.unit}"
                )

                return input_entity

        # Add a new attribute to the entity
        input_entity.attributes.append(self._result_to_input_data_attribute(result))

        return input_entity

    def add_results_to_input(
        self, data: InputDataModel, results: list[DataTransferComponentModel]
    ) -> InputDataModel:
        """
        Function to add the results of the components to the input data
        Args:
            data (InputDataModel): The input data to add the results to
            results (list[DataTransferComponentModel]): The results of the components
        Returns:
            InputDataModel: The input data with the results added
        """

        for result in results:

            for input_entity in data.input_entities:
                if input_entity.id == result.entity_id:

                    input_entity = self._add_result_to_input_entity(
                        result, input_entity
                    )
                    break

            # Add a new entity to the inputs
            data.input_entities.append(
                InputDataEntityModel(
                    id=result.entity_id,
                    attributes=[self._result_to_input_data_attribute(result)],
                )
            )

        return data

    async def calculation(self, data: InputDataModel) -> DataTransferModel:
        """
        Function to do the calculation

        Args:
            data (InputDataModel): Input data with the measured values for the calculation
        """

        all_component_results: list[DataTransferComponentModel] = []
        for component in self.components:
            try:
                component_results = component.run(data)

            except ValueError as e:
                logger.error(
                    f"Error occurred while running component "
                    f"{component.component_config.id}: {e}"
                )
                continue
            all_component_results.extend(component_results)
            if component != self.components[-1]:
                self.add_results_to_input(data, component_results)

        return DataTransferModel(components=all_component_results)

    async def calibration(self, data: InputDataModel) -> None:
        """
        Function to do the calibration of the component runner service.
        This function prepares the component with the static data, \
            if this is reloaded.
        It is possible to update the static data of the component with \
            rerunning the `prepare_start_component` method with new static data.

        Args:
            data (InputDataModel): InputDataModel for the component
        """
        logger.debug("The calibration of the components has begun.")
        for component in self.components:
            component.calibrate(
                static_data=data.static_entities if self.reload_staticdata else None
            )



"""
Description: Class to connect to a CrateDB and query data from it
Author: Martin Altenburger
"""

import crate.client
import pandas as pd
from filip.models.ngsi_v2.context import ContextEntity


class CrateDBConnection:
    """
    Class for a connection to a CrateDB

    Args:
        - crate_db_url: URL of the CrateDB
        - crate_db_user: Name of the User of CrateDB
        - crate_db_pw: Password of the User of CrateDB
        - crate_db_ssl: Verify the SSL-Cert?
    """

    def __init__(
        self,
        crate_db_url: str,
        crate_db_user: str = None,
        crate_db_pw: str = None,
        crate_db_ssl: bool = False,
    ) -> None:
        self.crate_db_url = crate_db_url
        self.crate_db_user = crate_db_user
        self.crate_db_pw = crate_db_pw
        self.crate_db_ssl = crate_db_ssl

    def get_database_connection(self) -> crate.client.connection:
        """
        Function to get a connection to the CrateDB
        Return:
            connection: Connection to the CrateDB
        """
        if self.crate_db_user is None:
            connection = crate.client.connect(self.crate_db_url)
        else:
            connection = crate.client.connect(
                self.crate_db_url,
                username=self.crate_db_user,
                password=self.crate_db_pw,
                verify_ssl_cert=self.crate_db_ssl,
            )

        return connection

    def query_existing_attributes(
        self,
        service: str,
        entity_type: str,
        attributes: list,
    ) -> list:
        """
        Function to query existing attributes of an entity type
        Args:
            service (str): Name of the Fiware Service
            entity_type (str): type of the entity
            attributes (list): list of attribute names
        Return:
            list: list of existing attributes
        """
        connection = self.get_database_connection()

        cursor = connection.cursor()

        attrs = "'time_index'"
        for attribute in list(attributes):
            attrs += ", '" + str(attribute) + "'"

        # check which column exists
        cursor.execute(
            f"SELECT column_name FROM information_schema.columns "
            f"WHERE table_name = 'et{entity_type}' "
            f"AND table_schema = 'mt{service}' "
            f"AND column_name IN ({attrs})"
        )
        attributes_db = list({column[0] for column in cursor.fetchall()})

        cursor.close()
        connection.close()

        return attributes_db

    def get_data(
        self,
        service: str,
        entity: ContextEntity,
        attributes: list,
        from_date: str,
        to_date: str,
        limit: int = 100000,
    ):
        """
        Function to query data from cratedb

        Args:
            - service: Name of the Fiware Service
            - entity (ContextEntity): Fiware Entity
            - attributes: list of attribute names
            - from_date: timestamp from which data is to be retrieved\
                (Milliseconds or Datetime (%Y-%m-%dT%H:%M:%S%z))
            - to_date: timestamp up to which data is to be retrieved\
                (Milliseconds or Datetime (%Y-%m-%dT%H:%M:%S%z))
            - limit: maximal number of datapoints

        Return:
            - dataframe with time index in utc and attributes as columns
        """

        attributes_db = self.query_existing_attributes(
            service=service, entity_type=entity.type, attributes=attributes
        )

        connection = self.get_database_connection()
        cursor = connection.cursor()

        # query existing columns
        attrs = ""
        attrs_not_null = ""
        for attribute in attributes_db:
            if attribute == attributes_db[-1]:
                attrs += '"' + str(attribute) + '"'

            else:
                attrs += '"' + str(attribute) + '", '

            if attribute != "time_index" and attrs_not_null != "":
                attrs_not_null += ' OR "' + str(attribute) + '" IS NOT NULL'
            elif attribute != "time_index":
                attrs_not_null += '"' + str(attribute) + '" IS NOT NULL'

        cursor.execute(
            f"SELECT {attrs} FROM mt{service}.et{entity.type} "
            f"WHERE entity_id = '{entity.id}' "
            f"AND time_index > '{from_date}' AND time_index < '{to_date}' "
            f"AND ({attrs_not_null}) "
            f"limit {limit}"
        )
        results = cursor.fetchall()

        if len(results) > 0:
            df = pd.DataFrame(results)

            df.columns = [desc[0] for desc in cursor.description]

            df.time_index = pd.to_datetime(df.time_index, unit="ms").dt.tz_localize(
                "UTC"
            )
            df.rename(columns={"time_index": "datetime"}, inplace=True)
            df.set_index(keys="datetime", drop=True, inplace=True)

        else:
            df = pd.DataFrame()

        cursor.close()
        connection.close()

        return df



"""
Modules for defining mediums used in thermal energy systems.
Author: Martin Altenburger, Paul Seidel
"""

from typing import Optional
from enum import Enum
from pydantic import BaseModel
from loguru import logger


class Medium(Enum):
    """
    Enum class for the mediums
    TODO: Add more mediums
    """

    WATER = "water"


class MediumParameters(BaseModel):
    """
    Base class for the medium parameters
    """

    cp: float
    rho: float


MEDIUM_VALUES = {Medium.WATER: MediumParameters(cp=4.19, rho=997)}


def get_medium_parameter(
    medium: Medium, temperature: Optional[float] = None
) -> MediumParameters:
    """Function to get the medium parameter
       - const values, if no temperature is set
       - calculation of cp and rho (constant pressure) as approximation of Glück
           - https://berndglueck.de/stoffwerte.php
           - "Zustandswerte Luft Wasser Dampf"  ISBN 3-345-00487-9

    Args:
        medium (Mediums): The medium
        temperature : float = None

    Returns:
        MediumParameters: The parameters of the medium
    """
    if medium not in MEDIUM_VALUES:
        logger.error(f"Medium {medium} not available")
        raise ValueError(f"Medium {medium} not available")

    if temperature is None:
        return MEDIUM_VALUES[medium]

    if medium is Medium.WATER:
        if temperature <= 0.1:
            logger.error("Attention! Temperature to low!")
            values = MediumParameters(cp=None, rho=None)
        elif temperature > 99.0:
            logger.error("Attention! Temperature to hight!")
            values = MediumParameters(cp=None, rho=None)
        else:
            # rho aproximation of Glück [kg/m³]:
            rho_calc = (
                1.002045 * 1000
                - 1.029905 * 0.1 * temperature
                - 3.698162 * 0.001 * temperature**2
                + 3.991053 * 0.000001 * temperature**3
            )
            # cp aproximation of Glück [kJ/kgK]:
            cp_calc = (
                4.177375
                - 2.144614 * 0.000001 * temperature
                - 3.165823 * 0.0000001 * temperature**2
                + 4.134309 * 0.00000001 * temperature**3
            )
            values = MediumParameters(cp=cp_calc, rho=rho_calc)

        return values
    # TODO add more mediums
    logger.error("Medium is not supported for temperature calculation")
    raise ValueError("Medium is not supported for temperature calculation")



"""
Description: Configuration models for the thermal storage component
Author: Martin Altenburger
"""

from typing import Optional
from enum import Enum
from pydantic import BaseModel, Field
from pydantic.functional_validators import model_validator
from encodapy.components.basic_component_config import (
    ComponentValidationError,
    InputData,
    OutputData,
    ConfigData,
)
from encodapy.utils.datapoints import DataPointGeneral, DataPointNumber, DataPointMedium
from encodapy.utils.mediums import Medium


class TemperatureLimits(BaseModel):
    """
    Configuration of the temperature limits in the termal storage, contains:
        - `minimal_temperature`: Minimal temperature in the thermal storage in °C
        - `maximal_temperature`: Maximal temperature in the thermal storage in °C
        - `reference_temperature`: Reference temperature in the storage in °C (default: 0°C)

    Raises:
        ValueError: if the minimal temperature is heighter than the maximal temperature
    """

    minimal_temperature: float = Field(
        ..., description="Minimal temperature in the thermal storage in °C"
    )
    maximal_temperature: float = Field(
        ..., description="Maximal temperature in the storage in °C"
    )
    reference_temperature: float = Field(
        0, description="Reference temperature in the storage in °C"
    )

    @model_validator(mode="after")
    def check_timerange_parameters(self) -> "TemperatureLimits":
        """Check the timerange parameters.

        Raises:
            ValueError: if the minimal temperature is heighter than the maximal temperature

        Returns:
            TemperatureLimits: The model with the validated parameters
        """

        if self.minimal_temperature > self.maximal_temperature:
            raise ValueError(
                "The minimal temperature should be lower than the maximal temperature"
            )

        return self


class StorageSensorConfig(BaseModel):
    """
    Configuration for the storage sensor in the thermal storage

    Contains:
        `height`: Height of the sensor in percent (0=top, 100=bottom)
        `limits`: Temperature limits for the sensor
    """

    height: float = Field(
        ...,
        ge=0,
        le=100,
        description="Height of the sensor in percent (0=top, 100=bottom)",
    )
    limits: TemperatureLimits


class ThermalStorageTemperatureSensors(BaseModel):
    """
    Configuration for the temperature sensors in the thermal storage

    Contains:
        `storage_sensors`: List of temperature sensors in the thermal storage

    It is required to set at least 3 sensors and no more than 10 sensors. The heights of the sensors
    must be between 0 and 100 percent and in ascending order.

    It is possible to add more information to sensors, thats the reason why this model is used.
    """

    storage_sensors: list[StorageSensorConfig] = Field(
        ..., description="List of temperature sensors (3–10 sensors)"
    )

    @model_validator(mode="after")
    def check_storage_tank_sensors(self) -> "ThermalStorageTemperatureSensors":
        """Check the storage tank sensors:
            - At least 3 sensors are required
            - No more than 10 sensors are allowed
            - Sensor heights must be between 0 and 100 percent
            - Sensor heights must be in ascending order

        Raises:
            ValueError: if the sensors are not set correctly

        Returns:
            ThermalStorageTemperatureSensors: The model with the validated parameters
        """

        if len(self.storage_sensors) < 3:
            raise ValueError("At least 3 storage sensors are required.")
        if len(self.storage_sensors) > 10:
            raise ValueError("No more than 10 storage sensors are allowed.")

        storage_sensor_height_ref = 0.0
        for _, storage_sensor in enumerate(self.storage_sensors):
            if storage_sensor.height < 0.0 or storage_sensor.height > 100.0:
                raise ValueError(
                    "Height of the sensor must be between 0 and 100 percent."
                )
            if storage_sensor.height < storage_sensor_height_ref:
                raise ValueError("Sensor heights must be in ascending order.")
            storage_sensor_height_ref = storage_sensor.height

        return self

    def get_number_of_sensors(self) -> int:
        """
        Get the number of storage sensors configured in the thermal storage.

        Returns:
            int: Number of storage sensors configured.
        """
        return len(self.storage_sensors)


class ThermalStorageInputData(InputData):
    """
    Model for the input of the thermal storage service, containing the temperature sensors
    in the thermal storage.

    The temperature sensors need to be set from 1 to 10, \
        no sensors are allowed to be missing between the others.

    Contains:
        `temperature_1` (DataPointNumber): first temperature sensor
        `temperature_2` (DataPointNumber): second temperature sensor
        `temperature_3` (DataPointNumber): third temperature sensor
        `temperature_4` (Optional[DataPointNumber]): fourth temperature sensor (optional)
        `temperature_5` (Optional[DataPointNumber]): fifth temperature sensor (optional)
        `temperature_6` (Optional[DataPointNumber]): sixth temperature sensor (optional)
        `temperature_7` (Optional[DataPointNumber]): seventh temperature sensor (optional)
        `temperature_8` (Optional[DataPointNumber]): eighth temperature sensor (optional)
        `temperature_9` (Optional[DataPointNumber]): ninth temperature sensor (optional)
        `temperature_10` (Optional[DataPointNumber]): tenth temperature sensor (optional)
        `temperature_in` (Optional[DataPointNumber]): consumer return temperature sensor \
            (optional)
        `temperature_out` (Optional[DataPointNumber]): consumer flow temperature sensor \
            (optional)
    """

    temperature_1: DataPointNumber = Field(
        ...,
        description="Input for the temperature of sensor 1 in the thermal storage",
        json_schema_extra={"unit": "CEL"},
    )
    temperature_2: DataPointNumber = Field(
        ...,
        description="Input for the temperature of sensor 2 in the thermal storage",
        json_schema_extra={"unit": "CEL"},
    )
    temperature_3: DataPointNumber = Field(
        ...,
        description="Input for the temperature of sensor 3 in the thermal storage",
        json_schema_extra={"unit": "CEL"},
    )
    temperature_4: Optional[DataPointNumber] = Field(
        None,
        description="Input for the temperature of sensor 4 in the thermal storage",
        json_schema_extra={"unit": "CEL"},
    )
    temperature_5: Optional[DataPointNumber] = Field(
        None,
        description="Input for the temperature of sensor 5 in the thermal storage",
        json_schema_extra={"unit": "CEL"},
    )
    temperature_6: Optional[DataPointNumber] = Field(
        None,
        description="Input for the temperature of sensor 6 in the thermal storage",
        json_schema_extra={"unit": "CEL"},
    )
    temperature_7: Optional[DataPointNumber] = Field(
        None,
        description="Input for the temperature of sensor 7 in the thermal storage",
        json_schema_extra={"unit": "CEL"},
    )
    temperature_8: Optional[DataPointNumber] = Field(
        None,
        description="Input for the temperature of sensor 8 in the thermal storage",
        json_schema_extra={"unit": "CEL"},
    )
    temperature_9: Optional[DataPointNumber] = Field(
        None,
        description="Input for the temperature of sensor 9 in the thermal storage",
        json_schema_extra={"unit": "CEL"},
    )
    temperature_10: Optional[DataPointNumber] = Field(
        None,
        description="Input for the temperature of sensor 10 in the thermal storage",
        json_schema_extra={"unit": "CEL"},
    )
    load_temperature_in: Optional[DataPointNumber] = Field(
        None,
        description="Input for the return temperature into the thermal storage (consumer)",
        json_schema_extra={"unit": "CEL"},
    )
    load_temperature_out: Optional[DataPointNumber] = Field(
        None,
        description="Input for the flow temperature from the thermal storage (consumer)",
        json_schema_extra={"unit": "CEL"},
    )

    @model_validator(mode="after")
    def check_storage_temperature_sensors(self) -> "ThermalStorageInputData":
        """
        Check that the storage sensors are configured.
        """
        previous_key = True
        for key, value in self.__dict__.items():
            if key.startswith("temperature") and value is None:
                previous_key = False
            if (
                key.startswith("temperature")
                and value is not None
                and previous_key is False
            ):
                raise ComponentValidationError(
                    f"Temperature sensor {key} is configured, "
                    "but the previous sensor is not configured. "
                    "Please check the configuration."
                )
        return self

    def check_load_connection_sensors(self) -> None:
        """
        Check if the load connection sensors are set

        Raises:
            ValueError: If any of the load connection sensors are not configured.
        """
        if self.load_temperature_in is None:
            raise ComponentValidationError(
                "Load temperature inflow sensor is not configured."
            )
        if self.load_temperature_out is None:
            raise ComponentValidationError(
                "Load temperature outflow sensor is not configured."
            )

    def get_number_storage_sensors(self) -> int:
        """
        Get the number of storage sensors configured in the thermal storage.

        Returns:
            int: Number of storage sensors configured.
        """
        return sum(
            1
            for key, value in self.__dict__.items()
            if key.startswith("temperature") and value is not None
        )


class ThermalStorageOutputData(OutputData):
    """
    Model for the output of the thermal storage service, containing the temperature sensors
    in the thermal storage.

    Each output data point is associated with a specific calculation method, \
        which is defined in the `json_schema_extra` field as `calculation`.

    Contains:
        `storage__level`: Optional[DataPointNumber] : Output for storage charge in percent \
            (0-100) (optional)
        `storage__energy`: Optional[DataPointNumber] : Output for storage energy in Wh (optional)
        `storage__loading_potential_nominal`: Optional[DataPointNumber] : \
            Output for storage loading potential in Wh (optional)
    """

    storage__level: Optional[DataPointNumber] = Field(
        None,
        description="Output for storage charge in percent (0-100)",
        json_schema_extra={"unit": "P1"},
    )
    storage__energy: Optional[DataPointNumber] = Field(
        None,
        description="Output for storage energy in Wh",
        json_schema_extra={"unit": "WHR"},
    )
    storage__loading_potential_nominal: Optional[DataPointNumber] = Field(
        None,
        description="Output for storage loading potential in Wh",
        json_schema_extra={"unit": "WHR"},
    )


class ThermalStorageCalculationMethods(Enum):
    """
    Enum for the calculation methods of the thermal storage service.

    Contains:
        - STATIC_LIMITS: Static limits given by the configuration
        - RETURN_LIMITS: Uses the temperature sensors from the in- and outflow as limits
    """

    STATIC_LIMITS = "static_limits"
    CONNECTION_LIMITS = "connection_limits"


class ThermalStorageEnergyTypes(Enum):
    """
    Enum for the energy types of the thermal storage service.
    
    Contains:
        Nominal ("nominal"): Nominal energy of the thermal storage \
            between the temperature limits
        Minimal ("minimal"): Minimal energy of the thermal storage \
            at the lower temperature limit
        Maximal ("maximal"): Maximal energy of the thermal storage \
            at the upper temperature limit
        Current ("current"): Current energy of the thermal storage \
            based on the current temperatures
    """

    NOMINAL = "nominal"
    MINIMAL = "minimal"
    MAXIMAL = "maximal"
    CURRENT = "current"


class DataPointCalculationMethod(DataPointGeneral):
    """
    Model for datapoints of the controller component which define the calculation method.

    Contains:
        value: The value of the datapoint, which is a string representing the calculation method
        unit: Optional unit of the datapoint, if applicable
        time: Optional timestamp of the datapoint, if applicable
    """

    value: ThermalStorageCalculationMethods


class DataPointSensorConfig(DataPointGeneral):
    """
    Model for datapoints of the controller component which define the sensor configuration.

    Contains:
        value: The value of the datapoint, which is a SensorConfig \
            representing the sensor configuration
        unit: Optional unit of the datapoint, if applicable
        time: Optional timestamp of the datapoint, if applicable
    """

    value: ThermalStorageTemperatureSensors


class ThermalStorageConfigData(ConfigData):
    """
    Model for the configuration data of the thermal storage service.
    
    Contains:
        `volume`: DataPointNumber : Volume of the thermal storage in m³
        `medium`: DataPointMedium : Medium of the thermal storage
        `sensor_config`: DataPointSensorConfig : \
            Sensor configuration of the thermal storage
        `calculation_method`: DataPointCalculationMethod : \
            Calculation method for the thermal storage
    """

    volume: DataPointNumber = Field(
        ...,
        description="Volume of the thermal storage in m³",
        json_schema_extra={"unit": "MTQ"},
    )
    medium: DataPointMedium = Field(
        DataPointMedium(value=Medium.WATER), description="Medium of the thermal storage"
    )
    sensor_config: DataPointSensorConfig = Field(
        ..., description="Sensor configuration of the thermal storage"
    )
    calculation_method: DataPointCalculationMethod = Field(
        DataPointCalculationMethod(
            value=ThermalStorageCalculationMethods.STATIC_LIMITS
        ),
        description="Calculation method for the thermal storage",
    )



"""
Simple Method to caluculate the energy in a the thermal storage
Author: Martin Altenburger, Paul Seidel
"""

from typing import Optional, Union

from loguru import logger

from encodapy.components.basic_component import BasicComponent
from encodapy.components.basic_component_config import (
    ComponentValidationError,
    ControllerComponentModel,
)
from encodapy.utils.datapoints import DataPointNumber
from encodapy.components.thermal_storage.thermal_storage_config import (
    TemperatureLimits,
    ThermalStorageCalculationMethods,
    ThermalStorageConfigData,
    ThermalStorageEnergyTypes,
    ThermalStorageInputData,
    ThermalStorageOutputData,
)
from encodapy.utils.mediums import get_medium_parameter
from encodapy.utils.models import (
    InputDataModel,
    StaticDataEntityModel,
)
from encodapy.utils.units import DataUnits


class ThermalStorage(BasicComponent):
    """
    Class to calculate the energy in a thermal storage.

    Service needs to be prepared before use (`prepare_start_thermal_storage`).

    Args:
        config (Union[ControllerComponentModel, list[ControllerComponentModel]]): \
            Configuration of the thermal storage
        static_data (Optional[list[StaticDataEntityModel]], optional): \
            Static data of the ThermalStorage
        component_id (str): ID of the thermal storage component
    
    """

    def __init__(
        self,
        config: Union[ControllerComponentModel, list[ControllerComponentModel]],
        component_id: str,
        static_data: Optional[list[StaticDataEntityModel]] = None,
    ) -> None:
        # Basic initialization of the thermal storage
        # Configuration of the thermal storage

        # Variables for the calculation
        self.sensor_volumes: Optional[dict] = None
        self.config_data: ThermalStorageConfigData
        self.input_data: ThermalStorageInputData
        self.output_data: ThermalStorageOutputData

        # Prepare Basic Parts / needs to be the latest part
        super().__init__(
            config=config, component_id=component_id, static_data=static_data
        )

    def _calculate_volume_per_sensor(self) -> dict:
        """
        Function to calculate the volume per sensor in the thermal storage

        Returns:
            dict: Volume per sensor in the thermal storage in m³
        """

        sensor_volumes = {}

        sensor_height_ref = 0.0

        for index, storage_sensor in enumerate(
            self.config_data.sensor_config.value.storage_sensors
        ):
            if index == len(self.config_data.sensor_config.value.storage_sensors) - 1:
                sensor_height_new = 100.0
            else:
                sensor_height_new = (
                    storage_sensor.height
                    + self.config_data.sensor_config.value.storage_sensors[
                        index + 1
                    ].height
                ) / 2

            sensor_volumes[index] = (
                (sensor_height_new - sensor_height_ref)
                / 100
                * self.config_data.volume.value
            )
            sensor_height_ref = sensor_height_new

        return sensor_volumes

    def _get_sensor_volume(self, sensor: int) -> float:
        """
        Function to get the volume of the sensors in the thermal storage

        Returns:
            float: Volume of the sensors in the thermal storage in m³
        """
        if self.sensor_volumes is None:
            raise ValueError("Sensor volumes are not set.")
        if sensor not in self.sensor_volumes:
            raise ValueError(f"Sensor {sensor} is not configured.")

        return round(self.sensor_volumes[sensor], 3)

    def _get_connection_limits(
        self, sensor_id: int, config_limits: TemperatureLimits
    ) -> TemperatureLimits:
        """
        Function to get the connection limits of the sensors in the thermal storage:
            - Uses the actual temperature of the outlet sensor (heat demand) as \
                minimal temperature of the upper storage temperature
            - Uses the actual temperature of the inlet sensor (heat demand) as \
                minimal temperature of the lower storage temperature
        Args:
            sensor_id (str): ID of the sensor in the thermal storage
        Returns:
            TemperatureLimits: Temperature limits of the sensors in the thermal storage
        """

        if self.input_data.load_temperature_out is None:
            logger.warning("Load temperature outflow is not set.")
            return config_limits
        if self.input_data.load_temperature_in is None:
            logger.warning("Load temperature inflow is not set.")
            return config_limits

        if sensor_id == 0:
            return TemperatureLimits(
                minimal_temperature=self.input_data.load_temperature_out.value,
                maximal_temperature=config_limits.maximal_temperature,
                reference_temperature=config_limits.reference_temperature,
            )

        return TemperatureLimits(
            minimal_temperature=self.input_data.load_temperature_in.value,
            maximal_temperature=config_limits.maximal_temperature,
            reference_temperature=config_limits.reference_temperature,
        )

    def _get_sensor_limits(self, sensor_id: int) -> TemperatureLimits:
        """
        Function to get the temperature limits of the sensors in the thermal storage
        Args:
            sensor (int): ID of the sensor in the thermal storage (0=top, 1=second, ...)
        Returns:
            TemperatureLimits: Temperature limits of the sensors in the thermal storage
        """

        config_limits = self.config_data.sensor_config.value.storage_sensors[
            sensor_id
        ].limits

        if (
            self.config_data.calculation_method.value
            == ThermalStorageCalculationMethods.STATIC_LIMITS
        ):
            return config_limits

        if (
            self.config_data.calculation_method.value
            == ThermalStorageCalculationMethods.CONNECTION_LIMITS
        ):
            limits = self._get_connection_limits(
                sensor_id=sensor_id, config_limits=config_limits
            )

            return limits

        logger.warning(
            f"Unknown calculation method: {self.config_data.calculation_method.value}"
        )

        return config_limits

    def get_storage_energy_content(
        self, energy_type: ThermalStorageEnergyTypes
    ) -> float:
        """
        Function to calculate the nominal energy content of the thermal storage

        Returns:
            float: Nominal energy content of the thermal storage in Wh
        """
        # Check if the calculation is possible

        if self.sensor_volumes is None:
            raise ValueError("Sensor volumes are not set.")

        nominal_energy = 0
        for index, _ in enumerate(self.config_data.sensor_config.value.storage_sensors):

            temperature_sensor = f"temperature_{index+1}"
            try:
                temperature: DataPointNumber = getattr(
                    self.input_data, temperature_sensor
                )
            except AttributeError as e:
                error_msg = (
                    f"Temperature sensor '{temperature_sensor}' "
                    "not found in input data."
                )
                logger.error(error_msg)
                raise AttributeError(error_msg) from e

            medium_parameter = get_medium_parameter(
                medium=self.config_data.medium.value,
                temperature=temperature.value,
            )

            sensor_limits = self._get_sensor_limits(sensor_id=index)

            if energy_type is ThermalStorageEnergyTypes.NOMINAL:
                temperature_difference = (
                    sensor_limits.maximal_temperature
                    - sensor_limits.minimal_temperature
                )

            elif energy_type is ThermalStorageEnergyTypes.MINIMAL:
                temperature_difference = (
                    sensor_limits.minimal_temperature
                    - sensor_limits.reference_temperature
                )

            elif energy_type is ThermalStorageEnergyTypes.MAXIMAL:
                temperature_difference = (
                    sensor_limits.maximal_temperature
                    - sensor_limits.reference_temperature
                )

            elif energy_type is ThermalStorageEnergyTypes.CURRENT:
                temperature_difference = (
                    temperature.value - sensor_limits.minimal_temperature
                )

            else:
                raise ValueError(f"Unknown energy type: {energy_type}")

            nominal_energy += (
                temperature_difference
                * self.sensor_volumes[index]
                * medium_parameter.rho
                * medium_parameter.cp
                / 3.6
            )

        return round(nominal_energy, 2)

    def get_storage_energy_nominal(self) -> tuple[float, DataUnits]:
        """
        Function to calculate the nominal energy content of the thermal storage

        Returns:
            tuple[float, DataUnits]: Nominal energy content of the thermal storage in Wh
        """

        return (
            self.get_storage_energy_content(ThermalStorageEnergyTypes.NOMINAL),
            DataUnits.WHR,
        )

    def get_storage_energy_minimum(self) -> tuple[float, DataUnits]:
        """
        Function to get the minimum energy content of the thermal storage

        Returns:
            tuple[float, DataUnits]: Minimum energy content of the thermal storage in Wh
        Raises:
            ValueError: If the thermal storage is not usable or the sensor values are not set
        """
        return (
            self.get_storage_energy_content(ThermalStorageEnergyTypes.MINIMAL),
            DataUnits.WHR,
        )

    def get_storage_energy_maximum(self) -> tuple[float, DataUnits]:
        """
        Function to get the maximum energy content of the thermal storage

        Returns:
            tuple[float, DataUnits]: Maximum energy content of the thermal storage in Wh
        Raises:
            ValueError: If the thermal storage is not usable or the sensor values are not set
        """
        return (
            self.get_storage_energy_content(ThermalStorageEnergyTypes.MAXIMAL),
            DataUnits.WHR,
        )

    def get_storage_energy_current(self) -> tuple[float, DataUnits]:
        """
        Function to get the current energy content of the thermal storage

        Returns:
            tuple[float, DataUnits]: Current energy content of the thermal storage in Wh
        Raises:
            ValueError: If the thermal storage is not usable or the sensor values are not set
        """
        return (
            self.get_storage_energy_content(ThermalStorageEnergyTypes.CURRENT),
            DataUnits.WHR,
        )

    def get_storage_loading_potential_nominal(self) -> tuple[float, DataUnits]:
        """
        Function to get the loading potential of the thermal storage, \
            which is the difference between the nominal and current energy content.

        Returns:
            tuple[float, DataUnits]: Loading potential of the thermal storage in Wh
        """
        nominal_energy = self.get_storage_energy_nominal()[0]
        current_energy = self.get_storage_energy_current()[0]
        loading_potential = round(nominal_energy - current_energy, 2)
        return loading_potential, DataUnits.WHR

    def set_input_data(self, input_data: InputDataModel) -> None:
        """
        Function to set the sensor values in the thermal storage

        Args:
            input_entities (list[InputDataEntityModel]): Input entities with temperature values
        Raises:
            ValueError: If the thermal storage is not usable or \
                the sensor values are not set correctly
        """

        super().set_input_data(input_data=input_data)
        if (
            self.config_data.calculation_method.value
            is ThermalStorageCalculationMethods.CONNECTION_LIMITS
        ):
            self.input_data.check_load_connection_sensors()

    def _check_temperatur_of_highest_sensor(self, state_of_charge: float) -> float:
        """
        Function to check if the temperature of the highest sensor is too low, \
            so there is no energy left
        Args:
            state_of_charge (float): Current state of charge

        Returns:
            float: Adjusted state of charge
        """

        temperature_limits = self._get_sensor_limits(sensor_id=0)
        ref_value = (
            temperature_limits.minimal_temperature
            + (
                temperature_limits.maximal_temperature
                - temperature_limits.minimal_temperature
            )
            * 0.1
        )

        if self.input_data.temperature_1.value < temperature_limits.minimal_temperature:
            return 0
        if self.input_data.temperature_1.value < ref_value:
            return (
                self.input_data.temperature_1.value
                - temperature_limits.minimal_temperature
            ) / (
                temperature_limits.maximal_temperature
                - temperature_limits.minimal_temperature
            )

        return state_of_charge

    def get_state_of_charge(self) -> tuple[float, DataUnits]:
        """
        Function to calculate the state of charge of the thermal storage

        If the temperature of the highest sensor is too low, there is no energy left, \
            so the state of charge is 0.

        Returns:
            tuple[float, DataUnits]: State of charge of the thermal storage in percent (0-100)
        """

        state_of_charge = (
            self.get_storage_energy_current()[0]
            / self.get_storage_energy_nominal()[0]
            * 100
        )

        state_of_charge = self._check_temperatur_of_highest_sensor(
            state_of_charge=state_of_charge
        )

        return round(state_of_charge, 2), DataUnits.PERCENT

    def get_storage__mean_temperature_maximal(self) -> DataPointNumber:
        """
        Function to calculate the mean maximal temperature of the thermal storage
        Using the maximal temperature of each sensor and \
        weighting it with the volume of the sensor

        Returns:
            DataPointNumber: Mean maximal temperature of the thermal storage in °C
        """
        max_temperatures: list[float] = []
        storage_volume = self.config_data.volume.value
        for index, storage_sensor in enumerate(
            self.config_data.sensor_config.value.storage_sensors
        ):

            sensor_volume = self._get_sensor_volume(sensor=index)

            max_temperatures.append(
                storage_sensor.limits.maximal_temperature * sensor_volume
            )

        return DataPointNumber(
            value=round(sum(max_temperatures) / storage_volume, 2),
            unit=DataUnits.DEGREECELSIUS,
        )

    def _check_input_configuration(self):
        """
        Function to check the input configuration of the service \
            in comparison to the sensor configuration.
        The inputs needs to match the sensor configuration.
        Raises:
            ValidationError: If the input configuration does not match the sensor configuration
        """

        if self.io_model is None:
            raise KeyError("No I/O model found in the thermal storage configuration.")

        inputs = ThermalStorageInputData.model_validate(self.io_model.input)
        # Check if there are all inputs avaiable
        if (
            self.config_data.calculation_method.value
            is ThermalStorageCalculationMethods.CONNECTION_LIMITS
        ):
            inputs.check_load_connection_sensors()

        # Check if all inputs are configured in the sensor configuration
        if inputs.get_number_storage_sensors() != len(
            self.config_data.sensor_config.value.storage_sensors
        ):
            raise ComponentValidationError(
                "Input configuration does not match sensor configuration."
                "Number of storage temperature sensors in config "
                f"({len(self.config_data.sensor_config.value.storage_sensors)}) "
                "is not the same like the number of inputs "
                f"({inputs.get_number_storage_sensors()})"
            )

    def prepare_component(self):
        """
        Function to prepare the thermal storage component for the start of the service
        """

        self._check_input_configuration()

        self.sensor_volumes = self._calculate_volume_per_sensor()

    def calculate(self):
        """
        Function to calculate the thermal storage values
        """
        storage__energy = self.get_storage_energy_current()
        storage__energy_datapoint = DataPointNumber(
            value=storage__energy[0],
            unit=storage__energy[1],
        )

        state_of_charge = self.get_state_of_charge()
        state_of_charge_datapoint = DataPointNumber(
            value=state_of_charge[0],
            unit=state_of_charge[1],
        )

        loading_potential = self.get_storage_loading_potential_nominal()
        loading_potential_datapoint = DataPointNumber(
            value=loading_potential[0],
            unit=loading_potential[1],
        )

        self.output_data = ThermalStorageOutputData(
            storage__energy=storage__energy_datapoint,
            storage__level=state_of_charge_datapoint,
            storage__loading_potential_nominal=loading_potential_datapoint,
        )



"""
Description: Configuration model for the two-point controller component.
Author: Martin Altenburger
"""

from pydantic import Field, model_validator
from loguru import logger
from encodapy.components.basic_component_config import (
    InputData,
    OutputData,
    ConfigData,
)
from encodapy.utils.datapoints import DataPointGeneral, DataPointNumber
from encodapy.utils.units import get_unit_adjustment_factor


class TwoPointControllerInputData(InputData):
    """
    Model for the input of the two-point controller component.
    
    Contains:
        current_value (DataPointNumber): The current value of the input.
        latest_control_signal (DataPointNumber): The latest control signal output \
            from the two-point controller.
    """

    current_value: DataPointNumber = Field(
        ..., description="Current value of the input, typically a sensor reading"
    )
    latest_control_signal: DataPointNumber = Field(
        ..., description="Latest control signal output from the two-point controller"
    )


class TwoPointControllerOutputData(OutputData):
    """
    Model for the output of the two-point controller component.

    Contains:
        control_signal (DataPointNumber): The control signal output from the two-point controller.
    """

    control_signal: DataPointGeneral = Field(
        ..., description="Control signal output from the two-point controller"
    )


class TwoPointControllerConfigData(ConfigData):
    """
    Model for the configuration data of the thermal storage service.
    """

    hysteresis: DataPointNumber = Field(
        ...,
        description="Hysteresis value for the two-point controller",
    )
    setpoint: DataPointNumber = Field(
        ...,
        description="Setpoint value for the two-point controller",
    )
    command_enabled: DataPointGeneral = Field(
        DataPointGeneral(value=1),
        description="Value representing the enabled state of the control signal",
    )
    command_disabled: DataPointGeneral = Field(
        DataPointGeneral(value=0),
        description="Value representing the disabled state of the control signal",
    )

    @model_validator(mode="after")
    def check_unit_setpoint(self):
        """
        Validator to check if the units of hysteresis and setpoint are the same.
        If not, it tries to convert the hysteresis to the unit of the setpoint.


        """
        hysteresis = DataPointNumber.model_validate(self.hysteresis)
        setpoint = DataPointNumber.model_validate(self.setpoint)
        if (
            hysteresis.unit != setpoint.unit
            and hysteresis.unit is not None
            and setpoint.unit is not None
        ):
            logger.warning(
                f"Units of hysteresis ({hysteresis.unit}) and setpoint ({setpoint.unit}) "
                "are not the same. Please check your configuration."
            )
            unit_adjustment_factor = get_unit_adjustment_factor(
                unit_actual=hysteresis.unit, unit_target=setpoint.unit
            )
            if unit_adjustment_factor is None:
                logger.warning(
                    f"Unit of hysteresis is {hysteresis.unit}, but expected {setpoint.unit}. "
                    "Could not convert, because units are not compatible "
                    "or no adjustment factor found."
                )
                return self

            hysteresis.value *= unit_adjustment_factor
            hysteresis.unit = setpoint.unit
        self.hysteresis = DataPointNumber(value=hysteresis.value, unit=hysteresis.unit)
        return self



"""
Description: This file contains the class FiwareConnections,
which is used to store the connection parameters for the Fiware and CrateDB connections.
Author: Paul Seidel
"""

import os
import json
import pathlib
from datetime import datetime
from typing import Union
from loguru import logger
import pandas as pd
from encodapy.config import (
    AttributeModel,
    AttributeTypes,
    CommandModel,
    DataQueryTypes,
    DefaultEnvVariables,
    FileExtensionTypes,
    InputModel,
    OutputModel,
    StaticDataModel,
    StaticDataFile,
)
from encodapy.utils.models import (
    InputDataAttributeModel,
    InputDataEntityModel,
    OutputDataEntityModel,
    StaticDataEntityModel,
)
from encodapy.utils.error_handling import NotSupportedError
from encodapy.utils.units import DataUnits


class FileConnection:
    """
    Class for the connection to a local file.
    Only a helper class.
    """

    def __init__(self):
        self.file_params = {}

    def load_file_params(self):
        """
        Function to load the file parameters
        """
        logger.debug("Load config for File interface")
        self.file_params["PATH_OF_INPUT_FILE"] = os.environ.get(
            "PATH_OF_INPUT_FILE", DefaultEnvVariables.PATH_OF_INPUT_FILE.value
        )
        self.file_params["START_TIME_FILE"] = os.environ.get(
            "START_TIME_FILE", DefaultEnvVariables.START_TIME_FILE.value
        )
        self.file_params["TIME_FORMAT_FILE"] = os.environ.get(
            "TIME_FORMAT_FILE", DefaultEnvVariables.TIME_FORMAT_FILE.value
        )
        self.file_params["PATH_OF_STATIC_DATA"] = os.environ.get(
            "PATH_OF_STATIC_DATA", DefaultEnvVariables.PATH_OF_STATIC_DATA.value
        )

    def _get_last_timestamp_for_file_output(
        self, output_entity: OutputModel
    ) -> tuple[OutputDataEntityModel, Union[datetime, None]]:
        """
        Function to get the latest timestamps of the output entity from a File, if exitst

        Args:
            output_entity (OutputModel): Output entity

        Returns:
            tuple[OutputDataEntityModel, Union[datetime, None]]:
                - OutputDataEntityModel with timestamps for the attributes
                - the latest timestamp of the output entity for the attribute
                with the oldest value (None if no timestamp is available)
        TODO:
            - is it really nessesary to get a timestamp for file-calculations /
            during calculation time is set to input_time
        """

        output_id = output_entity.id_interface

        timestamps = []
        timestamp_latest_output = None

        return (
            OutputDataEntityModel(id=output_id, attributes_status=timestamps),
            timestamp_latest_output,
        )

    def get_data_from_file(
        self,
        method: DataQueryTypes,
        entity: InputModel,
    ) -> Union[InputDataEntityModel, None]:
        """f
        Function to check input data-file and load data, \
        check of the file extension (compare in lower cases)

        Args:
            method (DataQueryTypes): Keyword for type of query
            entity (InputModel): Input entity

        Raises:
            NotSupportedError: If the file extension is not supported

        Returns:
            Union[InputDataEntityModel, None]: Model with the input data or \
                None if no data is available
        """

        file_extension = pathlib.Path(
            self.file_params["PATH_OF_INPUT_FILE"]
        ).suffix.lower()

        if file_extension == FileExtensionTypes.CSV.value:
            logger.debug(f"load inputdata from {file_extension} -file")
            data = self.get_data_from_csv_file(method=method, entity=entity)
        elif file_extension == FileExtensionTypes.JSON.value:
            logger.debug(f"load inputdata from {file_extension} -file")
            data = self.get_data_from_json_file(method=method, entity=entity)
        else:
            logger.debug(f"File extension {file_extension} is not supported")
            raise NotSupportedError

        return data

    def get_data_from_csv_file(
        self,
        method: DataQueryTypes,
        entity: InputModel,
    ) -> Union[InputDataEntityModel, None]:
        """
            Function to read input data for calculations from a input file.
            first step: read the first values in the file / id_inputs.
            Then get the data from the entity since the last timestamp
            of the output entity from cratedb.
        Args:
            - method (DataQueryTypes): Keyword for type of query
            - entity (InputModel): Input entity
        TODO:
             - timestamp_latest_output (datetime): Timestamp of the input value
             -  -> seperating Data in Calculation or here ??
             - handle the methods for the file interface

        Returns:
            - InputDataEntityModel: Model with the input data or None if the connection
            to the platform is not available

        """
        # TODO: Implement method handling for file interface
        _ = method  # Acknowledge unused parameter

        # attributes_timeseries = {}
        attributes_values = []
        path_of_file = self.file_params["PATH_OF_INPUT_FILE"]
        time_format = self.file_params["TIME_FORMAT_FILE"]
        try:
            data = pd.read_csv(path_of_file, parse_dates=["Time"], sep=";", decimal=",")
            data.set_index("Time", inplace=True)
            data.index = pd.to_datetime(data.index, format=time_format)
            # time = self.file_params["START_TIME_FILE"]
            # temp = data.loc[time, 'outside_Temperature']
        except FileNotFoundError:
            logger.error(f"Error: File not found ({path_of_file})")
            # TODO: What to do if the file is not found?
            return None
        for attribute in entity.attributes:

            if attribute.type == AttributeTypes.TIMESERIES:
                # attributes_timeseries[attribute.id] = attribute.id_interface
                logger.warning(
                    f"Attribute type {attribute.type} for attribute {attribute.id}"
                    f"of entity {entity.id} not supported."
                )
            elif attribute.type == AttributeTypes.VALUE:
                attributes_values.append(
                    InputDataAttributeModel(
                        id=attribute.id,
                        data=data[attribute.id_interface].iloc[0],
                        data_type=AttributeTypes.VALUE,
                        data_available=True,
                        latest_timestamp_input=data.index[0],
                    )
                )
            else:
                logger.warning(
                    f"Attribute type {attribute.type} for attribute {attribute.id}"
                    f"of entity {entity.id} not supported."
                )

        return InputDataEntityModel(id=entity.id, attributes=attributes_values)

    def get_data_from_json_file(
        self,
        method: DataQueryTypes,
        entity: InputModel,
    ) -> Union[InputDataEntityModel, None]:
        """
            Function to read input data for calculations from a input file.
            first step: read the keys and values in the file / id_inputs.
            Then get the data from the entity since the last timestamp
            of the output entity from cratedb.
        Args:
            - method (DataQueryTypes): Keyword for type of query
            - entity (InputModel): Input entity
        TODO:
             - timestamp_latest_output (datetime): Timestamp of the input value
             -  -> seperating Data in Calculation or here ??
             - handle the methods for the file interface

        Returns:
            - InputDataEntityModel: Model with the input data or None if the connection
            to the platform is not available

        """
        # TODO: Implement method handling for file interface
        _ = method  # Acknowledge unused parameter

        # attributes_timeseries = {}
        attributes_values = []
        path_of_file = self.file_params["PATH_OF_INPUT_FILE"]
        time_format = self.file_params["TIME_FORMAT_FILE"]
        try:
            # read data from json file and timestamp
            with open(path_of_file, encoding="utf-8") as f:
                data = json.load(f)
        except FileNotFoundError:
            logger.error(f"Error: File not found ({path_of_file})")
            # TODO: What to do if the file is not found?
            return None
        for attribute in entity.attributes:
            if attribute.type == AttributeTypes.TIMESERIES:
                # attributes_timeseries[attribute.id] = attribute.id_interface

                for input_data in data:
                    time = datetime.strptime(input_data["time"], time_format)
                    if attribute.id_interface == input_data["id_interface"]:
                        attributes_values.append(
                            InputDataAttributeModel(
                                id=attribute.id,
                                data=input_data["value"],
                                data_type=AttributeTypes.TIMESERIES,
                                data_available=True,
                                latest_timestamp_input=time,
                            )
                        )
            elif attribute.type == AttributeTypes.VALUE:

                attributes_values.append(
                    InputDataAttributeModel(
                        id=attribute.id,
                        data=data[attribute.id_interface].iloc[0],
                        data_type=AttributeTypes.VALUE,
                        data_available=True,
                        latest_timestamp_input=data.index[0],
                    )
                )
            else:
                logger.warning(
                    f"Attribute type {attribute.type} for attribute {attribute.id}"
                    f"of entity {entity.id} not supported."
                )

        return InputDataEntityModel(id=entity.id, attributes=attributes_values)

    def _get_unit_from_file(
        self,
        metadata: Union[dict[str, str], None],
    ) -> Union[DataUnits, None]:
        """
        Extracts the unit from the metadata dictionary.

        Args:
            metadata (Union[dict[str, str], None]): Metadata dictionary or None.
        Returns:
            Union[DataUnits, None]: Extracted data unit or None.
        """

        if metadata is None:
            return None
        if not isinstance(metadata, dict):
            logger.warning(f"Metadata is not a dictionary: {metadata}")
            return None
        metadata_lowercase = {k.lower(): v for k, v in metadata.items()}
        unit = metadata_lowercase.get("unitcode", None)
        if unit:
            return DataUnits(unit)
        return None

    def get_staticdata_from_file(
        self,
        entity: StaticDataModel,
    ) -> Union[StaticDataEntityModel, None]:
        """
        Function to read static data for calculations from config file.
        Args:
            - entity (StaticDataModel): Input entity
        TODO:
            - work with timeseries, for example: timetable with presence or heating_times

        Returns:
            - StaticDataEntityModel: Model with the static data

        """

        static_data_path = self.file_params["PATH_OF_STATIC_DATA"]

        attributes_values = []

        try:
            # read data from json file and timestamp
            with open(static_data_path, encoding="utf-8") as f:
                static_data = json.load(f)
        except FileNotFoundError:
            logger.error(f"Error: File not found ({static_data_path})")
            # TODO: What to do if the file is not found?
            return None
        if isinstance(static_data, list):
            static_data = {"staticdata": static_data}
        elif isinstance(static_data, dict):
            pass
        else:
            logger.error(f"Error: Unsupported data format ({static_data_path})")
            return None

        static_data = StaticDataFile.model_validate(static_data)

        for attribute in entity.attributes:

            for item_entity in static_data.staticdata:
                for item_attribute in item_entity.attributes:
                    if item_attribute.id == attribute.id:

                        attributes_values.append(
                            InputDataAttributeModel(
                                id=attribute.id,
                                data=item_attribute.value,
                                unit=self._get_unit_from_file(item_attribute.metadata),
                                data_type=AttributeTypes.VALUE,
                                data_available=True,
                                latest_timestamp_input=None,
                            )
                        )

        return StaticDataEntityModel(id=entity.id, attributes=attributes_values)

    def send_data_to_json_file(
        self,
        output_entity: OutputModel,
        output_attributes: list[AttributeModel],
        output_commands: list[CommandModel],
    ) -> None:
        """_Function to create a json_file in result-folder

        Args:
            output_entity (OutputModel): _description_
            output_attributes (list[AttributeModel]): _description_
            output_commands (list[CommandModel]): _description_

        Out: Json-file

        TODO:
            - Is it better to set the results-folder via env?
        """
        outputs = []
        commands = []
        logger.debug("Write outputs to json-output-files")

        if not os.path.exists("./results"):
            os.makedirs("./results")

        for output in output_attributes:
            outputs.append(
                {
                    "id_interface": output.id_interface,
                    "value": output.value,
                    "time": output.timestamp.isoformat(" "),
                }
            )

        with open(
            f"./results/outputs_{str(output_entity.id)}.json", "w", encoding="utf-8"
        ) as outputfile:
            json.dump(outputs, outputfile)

        for command in output_commands:
            commands.append(
                {
                    "id_interface": command.id_interface,
                    "value": command.value,
                    "time": command.timestamp.isoformat(" "),
                }
            )

        with open("./results/commands.json", "w", encoding="utf-8") as commandfile:
            json.dump(commands, commandfile)
